/*
 * Copyright (c) 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.audio;
import java.io.*;
import java.util.*;
import triad.backend.*;
import triad.console.*;

/**
 * The <code>IOAudio</code> class represents a method of interacting with the
 * user via speech input and output.
 */
public class IOAudio implements IOMethod
{
    /**
     * The method of getting audio input.
     */
    private InputMethod audioInput;

    /**
     * The method of generating audio output.
     */
    private OutputMethod audioOutput;

    /**
     * The language generator that creates the statements that will be output.
     */
    private LanguageGenerator language;

    /**
     * Whether or not to operate in careful mode, in which input is checked.
     */
    private boolean careful;

    /**
     * Last audio output.
     */
    private String lastAudio;

    /**
     * Whether the main menu has been displayed already.
     */
    private boolean mainDisplayed;

    /**
     * Whether or not player one will start the coming game.
     */
    private boolean p1Start;

    /**
     * The last console prompt output for the user.
     */
    private String lastPrompt;

    /**
     * The buffered reader of standard input, that is, the keyboard.
     */
    private BufferedReader in;

    /**
     * Creates a new <code>IOAudio</code> that will use the given input and
     * output methods.
     *
     * @param audioInput the method of getting audio input.
     * @param audioOutput the method of generating audio output.
     * @param language the generator that crafts the statements that will be
     * spoken.
     * @param careful whether or not to operate in careful mode, in which input
     * is checked.
     */
    public IOAudio(InputMethod audioInput, OutputMethod audioOutput,
		   LanguageGenerator language, boolean careful)
    {
	this.audioInput = audioInput;
	this.audioOutput = audioOutput;
	this.language = language;
	this.careful = careful;
        this.lastAudio = null;
        this.mainDisplayed = false;
        this.lastPrompt = null;
        this.in = new BufferedReader(new InputStreamReader(System.in));
    }

    /**
     * Prompts the user for whether or not they want to start a new game.
     *
     * @param predicted the hand that the Triple Triad Assistant believes that
     * the user will play.
     * @param bFactory the factory containing the current rules that will be
     * used to build a board.
     * @return the hand that the player will play, or <code>null</code> to
     * indicate that they wish to terminate the program.
     */
    public Hand gameStart(Hand predicted, BoardFactory bFactory)
    {
	HandBuilder hb = new HandBuilder(predicted);

        /* Output the initial card list */
        if (this.mainDisplayed)
            System.out.println();
        else
            this.mainDisplayed = true;
	this.printHandRules(hb, bFactory);

        /* Listen for a valid command */
        boolean valid = true;
        while (true) {
            if (valid)
                this.displayPrompt("New game | List cards | Correct # | " +
                                   "Quit: ");
            valid = true;
	    VoiceCommand command = this.getCommand();
	    VoiceCommand.Type type = command.getType();
            if (type == VoiceCommand.Type.NEW_GAME) {
                System.out.println("New game\n");
                this.audioInput.acknowledgeCommand();
		return hb.toHand();
	    }
	    else if (type == VoiceCommand.Type.QUIT) {
                System.out.println("Quit");
                this.audioInput.acknowledgeCommand();
                if (this.confirm(this.language.confirmQuit()) == true)
                    return null;
            }
            else if (type == VoiceCommand.Type.LIST_CARDS) {
                System.out.println("List cards");
                this.audioInput.acknowledgeCommand();
                this.sayAndRemember(this.language.cardList(hb));
            }
            else if (type == VoiceCommand.Type.CORRECT) {
		int toCorrect = command.getArgument(0);
                System.out.println("Correct " + toCorrect);
                this.audioInput.acknowledgeCommand();
                this.correctCard(hb, toCorrect-1);
            }
            else
                valid = false;
        }
    }

    /**
     * Returns a hand, input by the user, with which the opponent will play.
     *
     * @param opponent the opponent.
     * @return the hand that the opponent will play.
     */
    public Hand getOpponentHand(Player opponent)
    {
	/* Get the cards */
	HandBuilder hb = new HandBuilder();
        System.out.println("Hand for " + opponent);
	for (int i = 0; i < Hand.SIZE; i++)
	    hb.add(this.inputCard());

	/* If we're being careful, output the heard cards */
	if (this.careful)
	    this.sayAndRemember(this.language.cardList(hb));

        /* Listen for a valid command */
        boolean valid = true;
        while (true) {
            if (valid)
                this.displayPrompt("Player # starts | List cards | " +
                                   "Correct #: ");
            valid = true;
            VoiceCommand command = this.getCommand();
            VoiceCommand.Type type = command.getType();
            if (type == VoiceCommand.Type.START) {
                int firstToGo = command.getArgument(0);
                System.out.println("Player " + firstToGo + " starts\n");
                this.audioInput.acknowledgeCommand();
                this.p1Start = (firstToGo == 1);
		return hb.toHand();
	    }
            else if (type == VoiceCommand.Type.LIST_CARDS) {
                System.out.println("List cards");
                this.audioInput.acknowledgeCommand();
                this.sayAndRemember(this.language.cardList(hb));
            }
            else if (type == VoiceCommand.Type.CORRECT) {
		int toCorrect = command.getArgument(0);
                System.out.println("Correct " + toCorrect);
                this.audioInput.acknowledgeCommand();
                this.correctCard(hb, toCorrect-1);
            }
            else
                valid = false;
        }
    }

    /**
     * Returns which player is going first based on the user's input.
     *
     * @param playerOne player one.
     * @param playerTwo player two.
     * @return <code>true</code> if player one is going first, otherwise
     * <code>false</code>.
     */
    public boolean isPlayerOneFirst(Player playerOne, Player playerTwo)
    {
        return this.p1Start;
    }

    /**
     * Notifies the user that the given player's turn is beginning.
     *
     * @param player the player whose turn is beginning.
     */
    public void turnBegins(Player player)
    {
	/* Does nothing */
    }

    /**
     * Notifies the user of the final score of the game.
     *
     * @param playerOne player one.
     * @param playerTwo player two.
     * @param scoreOne the score for the first player.
     * @param scoreTwo the score for the second player.
     */
    public void finalScore(Player playerOne, Player playerTwo, int scoreOne,
                           int scoreTwo)
    {
	/* Does nothing */
    }

    /**
     * Notifies the user that an AI player is beginning a search to the given
     * depth.
     *
     * @param depth the depth of the search.
     */
    public void aiStartSearch(int depth)
    {
	this.sayAndRemember(this.language.startSearch(depth));
    }

    /**
     * Notifies the user that the AI's search from the previous call to
     * {@link #aiStartSearch} has finished.
     *
     * @param move the move that the AI chose to make.
     * @param aiScore the AI's projected score for itself.
     * @param oppScore the AI's projected score for its opponent.
     * @param evalRes the internal evaluation function result from the search.
     * @param timeMS the time the search took in milliseconds.
     */
    public void aiEndSearch(Move move, int aiScore, int oppScore, int evalRes,
                            long timeMS)
    {
	Card card = move.getCard();
	int position = move.getPosition();
	System.out.println("Play: " + card + " to spot " + position +
			   " (" + aiScore + " - " + oppScore + ")");
	this.sayAndRemember(this.language.aiPlay(card, position, aiScore,
						 oppScore));
    }

    /**
     * Prompts the user to input the card and position of a play.
     *
     * @param board the board which requires a play for the currently active
     * player.
     * @return the move that the user input, which could also be to concede or
     * abort the game.
     */
    public Move inputManualMove(Board board)
    {
       while (true) {
            this.displayPrompt("Play: ");
            Move move = this.inputValidMove(board);
            if (move.isAbort())
                System.out.println("Abort");
            else if (move.isConcede())
                System.out.println("Concede");
            else
		System.out.println(move.getCard().toString() + " to spot " +
				   move.getPosition());
            this.audioInput.acknowledgeCommand();

	    String conf = this.language.confirmPlay(move);
	    if (this.careful && this.confirm(conf) == false)
		continue;
	    return move;
       }
    }

    /**
     * Prompts the user to input the card and position of a play, verifying
     * the given move against a model.
     *
     * @param board the board which requires a play for the currently active
     * player.
     * @param sv the model against which to verify the input move.
     * @return the move that the user input, which could also be to concede or
     * abort the game.
     */
    public Move inputManualMove(Board board, SearchVerifier sv)
    {
	Move move;
	while (true) {
	    move = this.inputManualMove(board);
	    if (sv.verify(move))
		return move;
	    if (this.confirm(this.language.unexpectedMove()))
		return move;
	}
    }

    /**
     * Informs the player about the best opponent card to take.
     *
     * @param card the card to take, where <code>null</code> means that it does
     * not matter.
     */
    public void informBest(Card card)
    {
        String c = (card == null ? "anything" : card.toString());
        System.out.println("\nCard to take: " + c);
	this.sayAndRemember(this.language.informBest(card));
    }

    /**
     * Outputs the current set of cards in the hand builder and the current set
     * of rules in the board factory.
     *
     * @param hb the hand builder to output.
     * @param bf the board factory to output.
     */
    private void printHandRules(HandBuilder hb, BoardFactory bf)
    {
	ColumnDisplay cd = new ColumnDisplay(2);
	for (int i = 0; i < Hand.SIZE; i++)
	    cd.add(0, hb.getCard(i).toString());

	List<String> rules = bf.getRules();
	for (String rule : rules)
	    cd.add(1, rule);
	
	cd.display();
    }

    /**
     * Prompts the user to make a choice.
     *
     * @param prompt the prompt to display.
     */
    private void displayPrompt(String prompt)
    {
        System.out.print(prompt);
        System.out.flush();
        this.lastPrompt = prompt;
    }

    /**
     * Suspends the program until the user restarts it.
     */
    private void suspendProgram()
    {
        System.out.print("Suspended...");
        System.out.flush();
        this.audioOutput.say(this.language.suspend());

        String line = null;
        while (line == null) {
            try {
                line = this.in.readLine();
            }
            catch (IOException ioe) {
                System.out.println("Error reading enter from standard input");
            }
        }

        if (this.lastPrompt != null)
            this.displayPrompt(this.lastPrompt);
    }

    /**
     * Changes the card at the given index based on user input.
     *
     * @param hb the hand builder containing the cards.
     * @param index the index of the card that needs to be re-input.
     */
    private void correctCard(HandBuilder hb, int index)
    {
	this.sayAndRemember(this.language.correctCard(hb, index));
        Card newCard = this.inputCard();
        if (this.careful) {
	    String conf = this.language.confirmCorrectCard(hb, index, newCard);
            if (this.confirm(conf) == false)
                return;
        }
	hb.replace(index, newCard);
    }

    /**
     * Confirms with the user that the given question is correct.
     *
     * @param question the question that needs to be confirmed.
     * @return <code>true</code> if the user confirms the question, or
     * <code>false</code> if the question is rejected.
     */
    private boolean confirm(String question)
    {
        this.lastPrompt = null;
        this.sayAndRemember(question);
        while (true) {
            VoiceCommand command = this.getCommand();
            VoiceCommand.Type type = command.getType();
            if (type == VoiceCommand.Type.CONFIRM) {
                this.audioInput.acknowledgeCommand();
                return true;
            }
            else if (type == VoiceCommand.Type.REJECT) {
                System.out.println("  ^--> Cancelled");
                this.audioInput.acknowledgeCommand();
                return false;
            }
	}
    }

    /**
     * Listens on the microphone for a move that is currently valid.
     *
     * @param board the board on which the move would be played.
     * @return a move that is valid, including possible an abort or concede
     * move.
     */
    private Move inputValidMove(Board board)
    {
        while (true) {
            Move move = this.inputMove();
            if (move.isAbort() || move.isConcede())
                return move;

            Card card = move.getCard();
            int position = move.getPosition();
            if (board.isSquareEmpty(position-1) == false)
                continue;

	    for (int i = 0; i < Board.NUM_SIDE; i++) {
		Card inHand = board.cardFromSide(i);
		if (inHand != null && inHand.equals(card))
		    return move;
	    }
	}
    }

    /**
     * Listens on the microphone for a play command, valid or otherwise.
     *
     * @return the input move.
     */
    private Move inputMove()
    {
        while (true) {
            VoiceCommand command = this.getCommand();
            VoiceCommand.Type type = command.getType();
            if (type == VoiceCommand.Type.ABORT)
                return new Move(false);
            else if (type == VoiceCommand.Type.CONCEDE)
                return new Move(true);
            else if (type != VoiceCommand.Type.PLAY)
                continue;

            Card c;
            try {
                c = Card.get(command.getArgument(0), command.getArgument(1),
			     command.getArgument(2), command.getArgument(3));
            }
            catch (IllegalArgumentException iae) {
                continue;
            }
            return new Move(c, command.getArgument(4));
        }
    }

    /**
     * Listens on the microphone for a card command.
     *
     * @return the input card.
     */
    private Card inputCard()
    {
        this.displayPrompt("Card: ");
        while (true) {
            VoiceCommand command = this.getCommand();
            if (command.getType() != VoiceCommand.Type.CARD)
                continue;

            Card ret = null;
            try {
                ret = Card.get(command.getArgument(0), command.getArgument(1),
			       command.getArgument(2), command.getArgument(3));
            }
            catch (IllegalArgumentException iae) {
                continue;
            }

            System.out.println(ret);
	    this.audioInput.acknowledgeCommand();
	    return ret;
	}
    }

    /**
     * Captures audio input, acting on any commands that can be given at any
     * time, and returns a command other than an anytime command.
     *
     * @return a command to perform.
     */
    private VoiceCommand getCommand()
    {
	while (true) {
	    VoiceCommand vc = this.audioInput.getCommand();
	    VoiceCommand.Type type = vc.getType();
	    if (type == VoiceCommand.Type.SUSPEND) {
		this.audioInput.acknowledgeCommand();
		this.suspendProgram();
	    }
	    else if (type == VoiceCommand.Type.REPEAT) {
		if (this.lastAudio != null) {
		    this.audioInput.acknowledgeCommand();
		    this.audioOutput.say(this.lastAudio);
		}
	    }
	    else
		return vc;
        }
    }

    /**
     * Outputs the given audio, and remembers the last audio output, if the
     * given output string is non-empty.
     *
     * @param output the audio to output;
     */
    private void sayAndRemember(String output)
    {
        if (output.isEmpty())
            return;
        this.lastAudio = output;
        this.audioOutput.say(output);
    }
}
