/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.audio;
import edu.cmu.sphinx.frontend.util.Microphone;
import edu.cmu.sphinx.recognizer.Recognizer;
import edu.cmu.sphinx.result.Result;
import edu.cmu.sphinx.util.props.ConfigurationManager;
import javax.sound.sampled.*;
import java.io.*;
import java.util.*;

/**
 * The <code>InputSphinx4</code> class represents a way of getting voice input
 * using the Sphinx 4 voice recognition library as a backend.
 */
public class InputSphinx4 extends InputMethod
{
    /**
     * The different natural language inputs accepted for the numbers one
     * through ten.  We allow users to say "penta" for "five," because it
     * yields a more distinct sound.
     */
    private static final String[][] NUMBER_INPUTS =
    {
        {"one"},
        {"two"},
        {"three"},
        {"four"},
        {"five", "penta"},
        {"six"},
        {"seven"},
        {"eight"},
        {"nine"},
        {"ten"}
    };

    /**
     * The template file for the configuration file.
     */
    private static final String TEMPLATE_FILE =
        "../audio/tripleTriad.template.xml";

    /**
     * The keyword in the configuration template that has to be replaced with
     * the microphone index.
     */
    private static final String TEMPLATE_KEYWORD = "INSERT_MIXER_HERE";

    /**
     * The audio input configuration file.
     */
    private static final String CONFIG_FILE =
        "../audio/tripleTriad.config.xml";

    /**
     * The default microphone to use if the preferred one cannot be found.
     */
    private static final String DEFAULT_MICROPHONE = "default";

    /**
     * The microphone used for audio input.
     */
    private Microphone microphone;

    /**
     * The algorithm that translates the raw audio input into English words.
     */
    private Recognizer recognizer;

    /**
     * Creates a new <code>InputSphinx4</code> method, that will defer to the
     * given output method.
     *
     * @param output the audio output method.
     * @throws AudioException if the configuration file for Sphinx 4 could not
     * be generated.
     */
    public InputSphinx4(OutputMethod output)
    {
	this(output, null);
    }

    /**
     * Creates a new <code>InputSphinx4</code> method, that will defer to the
     * given output method.
     *
     * @param output the audio output method.
     * @param microphone the name of the microphone, as recognized by the Java
     * sound subsystem, to use; or <code>null</code> to use the default
     * microphone.
     * @throws AudioException if the configuration file for Sphinx 4 could not
     * be generated.
     */
    public InputSphinx4(OutputMethod output, String microphone)
    {
	super(output);

        try {
            InputSphinx4.generateConfigurationFile(microphone);
        }
        catch (IOException ioe) {
            throw new AudioException("Could not generate configuration file");
        }

        ConfigurationManager cm =
            new ConfigurationManager(InputSphinx4.CONFIG_FILE);
        this.microphone = (Microphone)cm.lookup("microphone");
        this.recognizer = (Recognizer)cm.lookup("recognizer");
        this.recognizer.allocate();
    }

    /**
     * Processes the user's speech input and returns a spoken command.
     *
     * @return a received voice command.
     */
    protected VoiceCommand inputSpeech()
    {
	/*
	 * The grammar presented to the Sphinx 4 backend highly restricts the
	 * types of input we can receive on the microphone, so we are free to
	 * rip apart this string and assume there are the right number of
	 * tokens and number-tokens appear where they should.
	 */
        StringTokenizer st = new StringTokenizer(this.getTextFromSpeech());
        String[] tokens = new String[st.countTokens()];
        for (int i = 0; i < tokens.length; i++)
            tokens[i] = st.nextToken();

	if (tokens[0].equals("repeat"))
	    return new VoiceCommand(VoiceCommand.Type.REPEAT);
	else if (tokens[0].equals("suspend"))
	    return new VoiceCommand(VoiceCommand.Type.SUSPEND);
	else if (tokens[0].equals("roger"))
	    return new VoiceCommand(VoiceCommand.Type.CONFIRM);
	else if (tokens[0].equals("negative"))
	    return new VoiceCommand(VoiceCommand.Type.REJECT);
	else if (tokens[0].equals("new"))
	    return new VoiceCommand(VoiceCommand.Type.NEW_GAME);
	else if (tokens[0].equals("quit"))
	    return new VoiceCommand(VoiceCommand.Type.QUIT);
	else if (tokens[0].equals("list"))
	    return new VoiceCommand(VoiceCommand.Type.LIST_CARDS);
	else if (tokens[0].equals("correct")) {
	    VoiceCommand vc = new VoiceCommand(VoiceCommand.Type.CORRECT);
	    vc.addArgument(this.toInt(tokens[1]));
	    return vc;
	}
	else if (tokens[0].equals("player")) {
	    VoiceCommand vc = new VoiceCommand(VoiceCommand.Type.START);
	    vc.addArgument(this.toInt(tokens[1]));
	    return vc;
	}
	else if (tokens[0].equals("card")) {
	    VoiceCommand vc = new VoiceCommand(VoiceCommand.Type.CARD);
	    vc.addArgument(this.toInt(tokens[1]));
	    vc.addArgument(this.toInt(tokens[2]));
	    vc.addArgument(this.toInt(tokens[3]));
	    vc.addArgument(this.toInt(tokens[4]));
	    return vc;
	}
	else if (tokens[0].equals("abort"))
	    return new VoiceCommand(VoiceCommand.Type.ABORT);
	else if (tokens[0].equals("concede"))
	    return new VoiceCommand(VoiceCommand.Type.CONCEDE);
	else if (tokens[0].equals("play")) {
	    VoiceCommand vc = new VoiceCommand(VoiceCommand.Type.PLAY);
	    vc.addArgument(this.toInt(tokens[1]));
	    vc.addArgument(this.toInt(tokens[2]));
	    vc.addArgument(this.toInt(tokens[3]));
	    vc.addArgument(this.toInt(tokens[4]));
	    vc.addArgument(this.toInt(tokens[7]));
	    return vc;
	}

	throw new AudioException("Unknown command received from Sphinx 4");
    }

    /**
     * Listens on the microphone for any recognized command.
     *
     * @return a recognized command input on the microphone.
     * @throws AudioException if the microphone cannot start listening.
     */
    private String getTextFromSpeech()
    {
        this.microphone.clear();
        if (this.microphone.startRecording() == false)
            throw new AudioException("Could not start microphone");

        String inputStr = null;
        while (true) {
            Result result = this.recognizer.recognize();
            if (result != null) {
                inputStr = result.getBestFinalResultNoFiller();
                if (inputStr.isEmpty() == false)
                    break;
            }
        }

        this.microphone.stopRecording();
	return inputStr;
    }

    /**
     * Returns the given spoken number, spelled out in English, as an integer.
     *
     * @param number the spoken number, spelled in English.
     * @return the number as an integer.
     */
    private int toInt(String number)
    {
        for (int i = 0; i < InputSphinx4.NUMBER_INPUTS.length; i++) {
            String[] knownInputs = InputSphinx4.NUMBER_INPUTS[i];
            for (int j = 0; j < knownInputs.length; j++) {
                if (number.equals(knownInputs[j]))
                    return i+1;
            }
        }
        throw new IllegalArgumentException("Not a known number: " + number);
    }

    /**
     * Generates the audio configuration file from the audio configuration
     * template.
     *
     * @param microphone the preferred microphone to use, or <code>null</code>
     * to use the default microphone.
     * @throws IOException on a read or write error.
     */
    private static void generateConfigurationFile(String microphone)
	throws IOException
    {
        BufferedReader in = null;
	BufferedWriter out = null;

        /*
         * The microphone to use is either "default" or a number representing a
         * mixer index, e.g., "2".
         */
        String microphoneString = InputSphinx4.DEFAULT_MICROPHONE;
	String output = microphoneString;
	if (microphone != null) {
	    int index = InputSphinx4.getMicrophoneIndex(microphone);
	    if (index == -1)
		System.out.println("Could not find microphone: " + microphone);
	    else {
		output = microphone;
		microphoneString = Integer.toString(index);
	    }
	}
	System.out.println("Using microphone: " + output + "\n");

        /*
         * Replace the placeholder with the index of the microphone to use (or
         * "default")
         */
	try {
	    in = new BufferedReader(new FileReader(InputSphinx4.TEMPLATE_FILE));
	    out = new BufferedWriter(new FileWriter(InputSphinx4.CONFIG_FILE));

	    boolean firstLine = true;
	    while (true) {
		String line = in.readLine();
		if (line == null)
		    break;
		if (firstLine)
		    firstLine = false;
		else
		    out.newLine();
		line = line.replace(InputSphinx4.TEMPLATE_KEYWORD,
				    microphoneString);
		out.write(line, 0, line.length());
	    }
	}
	finally {
	    if (in != null)
		in.close();
	    if (out != null)
		out.close();
	}
    }

    /**
     * Returns the index of the preferred microphone in the list of mixers.
     *
     * @param microphone the preferred microphone to use.
     * @return the index of the preferred microphone, or <code>-1</code> if it
     * cannot be found.
     */
    private static int getMicrophoneIndex(String microphone)
    {
        Mixer.Info[] mixerInfo = AudioSystem.getMixerInfo();
        for (int i = 0; i < mixerInfo.length; i++) {
            String name = mixerInfo[i].getName();
            if (name.equals(microphone))
                return i;
        }
        return -1;
    }

}
