/*
 * Copyright (c) 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.audio;
import triad.backend.*;

/**
 * The <code>LanguageFast</code> class generates statements for audio output
 * that are quick and to the point.
 */
public class LanguageFast implements LanguageGenerator
{
    /**
     * Creates a new <code>LanguageFast</code>.
     */
    public LanguageFast() {}

    /**
     * Returns a statement listing the cards in a hand.
     *
     * @return a statement listing the cards in a hand.
     */
    public String cardList(HandBuilder hb)
    {
	String out = "";
        for (int i = 0; i < hb.getNumCards(); i++) {
            if (i != 0)
                out += ", ";
            out += hb.getCard(i);
        }
	return out + ".";
    }

    /**
     * Returns a statement to output when the user wants to correct a card.
     *
     * @param hb the set of cards to be corrected.
     * @param index the index of the card to be corrected.
     * @return a statement acknowledging that the user is correcting a card.
     */
    public String correctCard(HandBuilder hb, int index)
    {
	return "";
    }

    /**
     * Returns a question confirming the selected change of card.
     *
     * @param hb the set of cards to be corrected.
     * @param index the index of the card to be corrected.
     * @param newCard the new card to place at that index.
     * @return a question of confirmation.
     */
    public String confirmCorrectCard(HandBuilder hb, int index, Card newCard)
    {
	return "Change card " + (index+1) + " into " + newCard + "?";
    }

    /**
     * Returns a statement that a search of the given depth is starting.
     *
     * @param depth the depth of the AI search.
     * @return a statement that a search is starting.
     */
    public String startSearch(int depth)
    {
	return "";
    }

    /**
     * Returns a statement informing the user of a chosen play.
     *
     * @param card the card to play.
     * @param position the position to which to play it.
     * @param aiScore the final score for the AI.
     * @param oppScore the final score for the opponent.
     * @return a statement about the play.
     */
    public String aiPlay(Card card, int position, int aiScore, int oppScore)
    {
        String score = Integer.toString(aiScore) + " " + oppScore;
        if (aiScore > oppScore)
            score += " win";
        else if (oppScore > aiScore)
            score += " loss";
        else
            score += " draw";
	return "Play: " + card + ", to spot " + position + ". " + score + ".";
    }

    /**
     * Returns a question asking the user to confirm the given move.
     *
     * @param move the move to confirm.
     * @return a question confirming the move.
     */
    public String confirmPlay(Move move)
    {
	if (move.isAbort())
	    return "Abort?";
	else if (move.isConcede())
	    return "Concede?";
	else
	    return move.getCard().toString() + ", spot " +
		move.getPosition() + "?";
    }

    /**
     * Informs the player about the best opponent card to take.
     *
     * @param card the card to take, where <code>null</code> means that it does
     * not matter.
     * @return the the instructions about which card to take.
     */
    public String informBest(Card card)
    {
	if (card == null)
	    return "Take anything.";
	else
	    return "Take " + card + ".";
    }

    /**
     * Returns a question asking the user if they wish to quit.
     *
     * @return a question asking the user if they wish to quit.
     */
    public String confirmQuit()
    {
	return "Really quit?";
    }

    /**
     * Returns an indication that the program is suspended.
     *
     * @return an indication that the program is suspended.
     */
    public String suspend()
    {
	return "Suspended.";
    }

    /**
     * Returns an indication that the program received an unexpected (as per
     * the opponent's model) move input.
     *
     * @return an indication of an unexpected move.
     */
    public String unexpectedMove()
    {
	return "Warning: move does not match model.  Proceed?";
    }
}
