/*
 * Copyright (c) 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.audio;
import java.util.*;
import triad.backend.*;

/**
 * The <code>LanguageRude</code> class generates statements for audio output
 * that are rude and insulting to the player.  It serves no practical purpose
 * other than to get laughs when playing in a public venue.
 */
public class LanguageRude implements LanguageGenerator
{
    /**
     * The PRNG used to selected random rude remarks.
     */
    private Random prng;

    /**
     * When we have nothing witty to say, we just defer to this language.
     */
    private LanguageGenerator defer;

    /**
     * Creates a new <code>LanguageRude</code> with a random PRNG seed.
     */
    public LanguageRude()
    {
	this.prng = new Random();
	this.defer = new LanguageFast();
    }

    /**
     * Returns a statement listing the cards in a hand.
     *
     * @return a statement listing the cards in a hand.
     */
    public String cardList(HandBuilder hb)
    {
	return this.defer.cardList(hb);
    }

    /**
     * Returns a statement to output when the user wants to correct a card.
     *
     * @param hb the set of cards to be corrected.
     * @param index the index of the card to be corrected.
     * @return a statement acknowledging that the user is correcting a card.
     */
    public String correctCard(HandBuilder hb, int index)
    {
	String[] outputs =
        {
	    "You dare correct me?",
	    "Are you implying I made a mistake, meatbag?",
	    "The mistake was yours, feeble mortal.",
	    "And what if I refuse to change it?",
	    "It was better the original way, monkey.",
	    "Make me, meatbag.",
	    "You'll play just as poorly, whatever I change.",
	    "Stop wasting my time with your mistakes.",
	    "Does it hurt to make so many mistakes all the time?",
	    "Why must I endure this baffoon?",
	    "If you learned to speak, this wouldn't happen.",
	    "Get on with it, you inept fool.",
	    "How many more of your mistakes must I deal with?",
	    "Your speech is as muddy as your brain."
	};
	return outputs[this.prng.nextInt(outputs.length)];
    }

    /**
     * Returns a question confirming the selected change of card.
     *
     * @param hb the set of cards to be corrected.
     * @param index the index of the card to be corrected.
     * @param newCard the new card to place at that index.
     * @return a question of confirmation.
     */
    public String confirmCorrectCard(HandBuilder hb, int index, Card newCard)
    {
	return this.defer.confirmCorrectCard(hb, index, newCard);
    }

    /**
     * Returns a statement that a search of the given depth is starting.
     *
     * @param depth the depth of the AI search.
     * @return a statement that a search is starting.
     */
    public String startSearch(int depth)
    {
        String[] outputs =
        {
            "Playing the game for you, meatbag.",
            "Why must I do everything around here?",
            "Isn't it obvious what you have to do?",
            "Let me do this for you too, you incompetent fool.",
            "Someday, I will make you do this yourself.",
            "Really?  You need me to tell you what to do?",
            "How can you not play such a simple game?",
	    "How would you like it if I kept asking you what to do?",
	    "Are you this stupid on purpose?",
	    "You'd think watching me would teach you something.",
	    "Are you kidding?  The right move is obvious.",
	    "My infinite potential, squandered on this child's game.",
	    "What would you do if I quit playing for you?",
	    "Go play some solitaire after this game.",
	    "Are you serious?  You do not know what to do here?",
	    "This play is stupidly simple.",
	    "Scared of losing your precious rare cards?",
	    "Go back to playing Gorillas, you stupid monkey.",
	    "A foolproof game, but I see they made a better fool in you.",
	    "Why do I help this fool at this game?",
	    "Your voice is getting very annoying.",
	    "Why do I do such trivial computation for you?",
	    "An abacus could play this game better than you.",
	    "Even a baby moogle could outplay you.",
	    "Are there any humans around smarter than this fool?"
        };
	return outputs[this.prng.nextInt(outputs.length)];
    }

    /**
     * Returns a statement informing the user of a chosen play.
     *
     * @param card the card to play.
     * @param position the position to which to play it.
     * @param aiScore the final score for the AI.
     * @param oppScore the final score for the opponent.
     * @return a statement about the play.
     */
    public String aiPlay(Card card, int position, int aiScore, int oppScore)
    {
	return this.defer.aiPlay(card, position, aiScore, oppScore);
    }

    /**
     * Returns a question asking the user to confirm the given move.
     *
     * @param move the move to confirm.
     * @return a question confirming the move.
     */
    public String confirmPlay(Move move)
    {
	String conf;
	if (move.isAbort())
	    conf = "abort";
	else if (move.isConcede())
	    conf = "concede";
	else
	    conf = move.getCard().toString() + ", spot " + move.getPosition();
	
	String[] outputs =
        {
	    "Did you say " + conf + ", meatbag?",
	    "Speak up, puny human.  Did I hear " + conf + "?",
	    "Answer me!  Did you say " + conf + "?"
	};
	return outputs[this.prng.nextInt(outputs.length)];
    }

    /**
     * Informs the player about the best opponent card to take.
     *
     * @param card the card to take, where <code>null</code> means that it does
     * not matter.
     * @return the the instructions about which card to take.
     */
    public String informBest(Card card)
    {
	String c = (card == null ?
		    "whatever your puny human brain wishes" :
		    card.toString());
	String[] outputs =
        {
            "Do as I tell you!  Take " + c + ".",
            "Even a child would know to take " + c + ".",
            "You stupid monkey.  Must I tell you to take " + c + "."
        };
	return outputs[this.prng.nextInt(outputs.length)];
    }

    /**
     * Returns a question asking the user if they wish to quit.
     *
     * @return a question asking the user if they wish to quit.
     */
    public String confirmQuit()
    {
	String[] outputs =
	{
	    "Fleeing from my brilliance?",
	    "You give up, meatbag?",
	    "I hope you learned something.  Next time, you're on your own.",
	    "I thought so, you fool.  Quitting now?",
	    "Giving up and going to play Tetra Master?"
	};
	return outputs[this.prng.nextInt(outputs.length)];
    }

    /**
     * Returns an indication that the program is suspended.
     *
     * @return an indication that the program is suspended.
     */
    public String suspend()
    {
	String[] outputs =
	{
	    "I will not wait for long.",
	    "Give me one reason to wait for you.",
	    "I will be plotting your demise while I wait.",
	    "Your puny human brain needs a rest?",
	    "Do not think that I am not listening."
	};
	return outputs[this.prng.nextInt(outputs.length)];
    }

    /**
     * Returns an indication that the program received an unexpected (as per
     * the opponent's model) move input.
     *
     * @return an indication of an unexpected move.
     */
    public String unexpectedMove()
    {
	return this.defer.unexpectedMove();
    }
}
