/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.audio;
import java.util.*;

/**
 * The <code>OutputMethod</code> class represents a method of generating speech
 * output.
 */
public abstract class OutputMethod
{
    /**
     * The set of queued statements.
     */
    private Queue<String> queue;

    /**
     * If audio is queued to be output or is currently being output.
     */
    private boolean isOutputting;

    /**
     * The thread responsible for outputting all of the speech.
     */
    private Thread speechThread;

    /**
     * Creates a new <code>OutputMethod</code> with no statements yet queued to
     * speak.
     */
    public OutputMethod()
    {
	this.queue = new LinkedList<String>();
	this.isOutputting = false;
	this.speechThread = new OutputThread();
	this.speechThread.start();
    }

    /**
     * Queues the given statement to be output via audio.  Output happens
     * asynchronously, and this method can return before audio output has even
     * begun, let alone finished.  If the argument is the empty string, this
     * function does nothing.
     *
     * @param statement the statement to say.
     */
    public void say(String statement)
    {
	if (statement.isEmpty())
	    return;
	synchronized (this.queue) {
	    this.queue.add(statement);
	    this.isOutputting = true;
	    this.queue.notifyAll();
	}
    }

    /**
     * Returns once all of the queued statements have been completely spoken.
     */
    public void waitForComplete()
    {
	synchronized (this.queue) {
	    while (this.isOutputting) {
		try {
		    this.queue.wait();
		}
		catch (InterruptedException iex) {}
	    }
	}
    }

    /**
     * Outputs the given message in speech, returning when the audio output is
     * complete.
     *
     * @param message the message to output.
     * @throws AudioException if there is an error outputting the speech.
     */
    protected abstract void outputSpeech(String message);

    /**
     * The <code>OutputThread</code> class is responsible for running
     * asynchronously to whatever class is using the <code>OutputMethod</code>
     * class and outputting the audio messages in order.
     */
    private class OutputThread extends Thread
    {
	/**
	 * Creates a new <code>OutputThread</code> that will poll the queue for
	 * messages to output.
	 */
	public OutputThread()
	{
	    super();
	    this.setDaemon(true);
	}

	/**
	 * Polls the queue for messages to output and outputs them.
	 */
	public void run()
	{
	    while (true) {
		String toOutput;

		/* Get the next thing to output from the queue */
		synchronized (queue) {
		    while (queue.isEmpty()) {
			isOutputting = false;
			queue.notifyAll();
			try {
			    queue.wait();
			}
			catch (InterruptedException iex) {}
		    }
		    toOutput = queue.remove();
		}

		/* Output it as speech */
		outputSpeech(toOutput);
	    }
	}
    }
}
