/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.audio;
import java.util.*;

/**
 * The <code>VoiceCommand</code> class represents a single command given to an
 * {@link IOAudio} user interface.
 */
public class VoiceCommand
{
    /**
     * The different command types that the user can dictate.
     */
    public enum Type
    {
	/**
	 * A command for the audio output system to repeat the last thing it
	 * said.
	 */
	REPEAT,

	/**
	 * A command for the audio input to be suspended until a key press.
	 */
	SUSPEND,

	/**
	 * A command to answer in the affirmative to a question asked by the
	 * audio output system.
	 */
	CONFIRM,

	/**
	 * A command to answer in the negative to a question asked by the audio
	 * output system.
	 */
        REJECT,

	/**
	 * A command to start a new game.
	 */
        NEW_GAME,

        /**
	 * A command to quit the program.
	 */
	QUIT,

	/**
	 * A command to list the cards with which a player is playing.
	 */
	LIST_CARDS,

	/**
	 * A command to correct one of the listed cards.  It has one numerical
	 * argument, representing which card (indexed from <code>1</code>) to
	 * correct.
	 */
	CORRECT,

	/**
	 * A command to start the game with the previously indicated cards.  It
	 * has one numerical argument, representing which player (one or two)
	 * moves first.
	 */
	START,

	/**
	 * A command to input a card where card input is required.  It has four
	 * numerical arguments, representing the card edges.
	 */
	CARD,

	/**
	 * A command to abort the match.
	 */
        ABORT,

	/**
	 * A command to concede the match.
	 */
        CONCEDE,

	/**
	 * A command to play a card during the match.  It has five numerical
	 * arguments: the four edges, and the spot (indexed from
	 * <code>1</code>) to which to play the card.
	 */
        PLAY
    }

    /**
     * The type of command.
     */
    private Type ct;

    /**
     * The numerical arguments that have been given to the command.
     */
    private List<Integer> args;

    /**
     * Creates a new <code>VoiceCommand</code> of the given type.
     *
     * @param ct the type of command to create.
     */
    public VoiceCommand(Type ct)
    {
	this.ct = ct;
	this.args = new ArrayList<Integer>();
    }

    /**
     * Adds the given argument to the voice command.
     *
     * @param arg the argument to add.
     */
    public void addArgument(int arg)
    {
	this.args.add(arg);
    }

    /**
     * Returns the argument at the given index.
     *
     * @param index the index of the argument to return.
     * @throws ArrayIndexOutOfBoundsException if the index is negative, or
     * greater than or equal to the number of arguments that have been added.
     */
    public int getArgument(int index)
    {
	return this.args.get(index);
    }

    /**
     * Returns the command type of this command.
     *
     * @return the type of command this command represents.
     */
    public Type getType()
    {
	return this.ct;
    }
}
