/*
 * Copyright (c) 2010, 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>Board</code> class represents the nine-square board and two side
 * decks of the players in Triple Triad.
 */
public abstract class Board
{
    /**
     * The number of locations on the Triple Triad playing area, indexed from
     * <code>0</code> to <code>NUM_SQUARES - 1</code>.
     */
    public static final int NUM_SQUARES = 9;

    /**
     * The number of locations in each player's size deck, indexed from
     * <code>0</code> to <code>NUM_SIDE - 1</code>.
     */
    public static final int NUM_SIDE = Hand.SIZE;

    /**
     * By default, player one is considered the active player in this class'
     * methods.  However, if <code>amPlayerOne</code> is <code>false</code>,
     * then player two is the active player.
     */
    protected boolean amPlayerOne;

    /**
     * The positions on the playing board, indexed from <code>0</code> to
     * <code>NUM_SQUARES - 1</code>.
     */
    protected Card[] board;

    /**
     * If <code>board[i]</code> is not <code>null</code>, this array denotes if
     * the card at that location is owned by player one.  If it is
     * <code>null</code>, the value in this array is undefined.
     */
    protected boolean[] ownedByOne;

    /**
     * The side cards of player one.
     */
    private Card[] sideOne;

    /**
     * The side cards of player two.
     */
    private Card[] sideTwo;

    /**
     * The number of cards that have been played onto the board.
     */
    private int movesMade;

    /**
     * The total number of cards owned by player one.
     */
    private int scoreOne;

    /**
     * The position in <code>0</code> to <code>NUM_SQUARES - 1</code> where
     * the card was played on move <code>i</code>, where <code>i</code> is in
     * <code>0</code> to <code>NUM_SQUARES - 1</code>.  If that move has not
     * yet been played, the value is undefined.
     */
    private int[] moveLocation;

    /**
     * The index in <code>0</code> to <code>NUM_SIDE - 1</code> from which the
     * card was played on move <code>i</code>, where <code>i</code> is in
     * <code>0</code> to <code>NUM_SQUARES - 1</code>.  If that move has not
     * yet been played, the value is undefined.
     */
    private int[] moveCardIndex;

    /**
     * The entry at index <code>i</code> represents the number of cards flipped
     * by the play of move <code>i</code>.
     */
    private int[] moveNumFlipped;

    /**
     * The value at <code>cardsFlipped[i][j]</code> represents the
     * <code>j<code>th location, in <code>0</code> to
     * <code>NUM_SQUARES - 1</code>, flipped when move <code>i</code> was
     * played.
     */
    private int[][] moveCardsFlipped;

    /**
     * Creates a new empty Triple Triad <code>Board</code> in which the two
     * players play with the given hands.
     *
     * @param handOne the starting hand for player one.
     * @param handTwo the starting hand for player two.
     * @param maxFlipsPerMove an upper bound on the number of cards that can be
     * flipped in any single move.
     */
    public Board(Hand handOne, Hand handTwo, int maxFlipsPerMove)
    {
	/* Scorekeeping */
	this.amPlayerOne = true;
	this.movesMade = 0;
	this.scoreOne = Hand.SIZE;

	/* State of the main play area and side-decks */
        this.board = new Card[Board.NUM_SQUARES];
        this.ownedByOne = new boolean[Board.NUM_SQUARES];
	this.sideOne = new Card[Board.NUM_SIDE];
	this.sideTwo = new Card[Board.NUM_SIDE];
        for (int i = 0; i < Board.NUM_SQUARES; i++)
            this.board[i] = null;
	for (int i = 0; i < Board.NUM_SIDE; i++) {
	    this.sideOne[i] = handOne.getCard(i);
	    this.sideTwo[i] = handTwo.getCard(i);
	}

	/* Undo functionality */
	this.moveLocation = new int[Board.NUM_SQUARES];
	this.moveCardIndex = new int[Board.NUM_SQUARES];
	this.moveNumFlipped = new int[Board.NUM_SQUARES];
	this.moveCardsFlipped = new int[Board.NUM_SQUARES][];
        for (int i = 0; i < this.moveCardsFlipped.length; i++)
            this.moveCardsFlipped[i] = new int[maxFlipsPerMove];
    }

    /**
     * By default, player one is considered the active player, but each call to
     * <code>invert</code> changes which player is the active player.
     */
    public void invert()
    {
	this.amPlayerOne = !(this.amPlayerOne);
    }

    /**
     * Returns the number of free locations still available for play on the
     * main area of the board.
     *
     * @return the number of available locations to play.
     */
    public int numFreeLocations()
    {
	return Board.NUM_SQUARES - this.movesMade;
    }

    /**
     * Returns the total number of cards that the active player owns, both in
     * the field of play and remaining in their side-deck.
     *
     * @return the total number of cards that the active player owns.
     */
    public int score()
    {
	return (this.amPlayerOne ?
		this.scoreOne :
		2 * Hand.SIZE - this.scoreOne);
    }

    /**
     * Returns the card at the given index from the active player's side-deck,
     * or <code>null</code> if it has been moved into the field of play.
     *
     * @param index the index into the side-deck from which to get a card.
     * @return the card at the given index.
     */
    public Card cardFromSide(int index)
    {
	return (this.amPlayerOne ?
		this.sideOne[index] :
		this.sideTwo[index]);
    }

    /**
     * Returns whether the space at the given index, indexed from
     * <code>0</code> to <code>NUM_SQUARES - 1</code>, is empty and available
     * for play.
     *
     * @param index the index of the board space.
     * @return <code>true</code> if the space is empty, otherwise
     * <code>false</code>.
     */
    public boolean isSquareEmpty(int index)
    {
	return (this.board[index] == null);
    }

    /**
     * Returns whether or not the given card in the active player's side-deck
     * is still available.
     *
     * @param index the index of the side-deck space.
     * @return <code>true</code> if the card at the given index in the
     * side-deck is still available to be played, or <code>false</code> if it
     * has already been played into the main playing area.
     */
    public boolean isCardOnSide(int index)
    {
	return (this.amPlayerOne ?
		(this.sideOne[index] != null) :
		(this.sideTwo[index] != null));
    }

    /**
     * Plays the given card from my hand to the given space on the board.
     *
     * @param move the move to play.
     * @throws IllegalArgumentException if the card is not available to play or
     * if the board space is already occupied, or if the given move is to abort
     * or concede the match.
     */
    public void play(Move move)
    {
	if (move.isAbort() || move.isConcede())
	    throw new IllegalArgumentException("Unsupported move");
	Card moveCard = move.getCard();

	Card[] side = (this.amPlayerOne ? this.sideOne : this.sideTwo);
	for (int i = 0; i < side.length; i++) {
	    if (side[i] != null && side[i].equals(moveCard)) {
		this.play(i, move.getPosition()-1);
		return;
	    }
	}

	throw new IllegalArgumentException("Card not available to play");
    }

    /**
     * Plays the card at the given index in my hand to the given space on the
     * board.
     *
     * @param cardIndex the index into my hand at which to grab the card to
     * play.
     * @param spaceIndex the space on the board onto which to play the card.
     * @throws IllegalArgumentException if the card has already been played or
     * if the board space is already occupied.
     */
    public void play(int cardIndex, int spaceIndex)
    {
	/* Ensure card and spot are legal */
	Card[] side = (this.amPlayerOne ? this.sideOne : this.sideTwo);
	if (side[cardIndex] == null)
	    throw new IllegalArgumentException("Card not available for play");
	if (this.board[spaceIndex] != null)
	    throw new IllegalArgumentException("Space not available for play");

	/* Place the card on the board */
	this.board[spaceIndex] = side[cardIndex];
	this.ownedByOne[spaceIndex] = this.amPlayerOne;
	side[cardIndex] = null;

	/* Compute card flips */
	int[] flipList = this.moveCardsFlipped[this.movesMade];
	int numFlipped = this.computeFlips(spaceIndex, flipList);
	for (int i = 0; i < numFlipped; i++)
	    this.ownedByOne[flipList[i]] = this.amPlayerOne;

	/* Update score */
	if (this.amPlayerOne)
	    this.scoreOne += numFlipped;
	else
	    this.scoreOne -= numFlipped;

	/* Store undo information and increment move */
	this.moveLocation[this.movesMade] = spaceIndex;
	this.moveCardIndex[this.movesMade] = cardIndex;
	this.moveNumFlipped[this.movesMade] = numFlipped;
	this.movesMade++;
    }

    /**
     * Undoes the most recent call to {@link #play} that has not yet been
     * undone.
     *
     * @throws IllegalStateException if there are no calls to <code>play</code>
     * that have not been undone.
     */
    public void undo()
    {
	/* Decrement the moves counter */
	if (this.movesMade == 0)
	    throw new IllegalStateException("No cards in play");
	this.movesMade--;

	/* Undo all of the flipped cards */
        int numFlipped = this.moveNumFlipped[this.movesMade];
        for (int i = 0; i < numFlipped; i++) {
            int flippedLoc = this.moveCardsFlipped[this.movesMade][i];
            this.ownedByOne[flippedLoc] = !(this.ownedByOne[flippedLoc]);
        }

        /* Return the card to the hand of whomever played it, update score */
        int location = this.moveLocation[this.movesMade];
	int cardIndex = this.moveCardIndex[this.movesMade];
        if (this.ownedByOne[location]) {
            this.sideOne[cardIndex] = this.board[location];
            this.scoreOne -= numFlipped;
        }
        else {
            this.sideTwo[cardIndex] = this.board[location];
            this.scoreOne += numFlipped;
        }
        this.board[location] = null;
    }

    /**
     * Computes all of the cards flipped to the current player's colour on a
     * given play.  The current player (denoted by {@link #amPlayerOne}) just
     * played the card to the given index (the card can be seen via
     * {@link #board}).  This method computes all of the card indices that are
     * flipped from their current owner (stored in {@link #ownedByOne}) to the
     * current player, and stores all of the flipped card indices in the given
     * list.  Note that this method is not responsible for updating the
     * <code>ownedByOne</code> array.
     *
     * @param space the positional index on the board of the card just played
     * by the current player.
     * @param flipList the array to be filled with the list of flipped board
     * indices.
     * @return the number of cards flipped.
     */
    protected abstract int computeFlips(int space, int[] flipList);
}