/*
 * Copyright (c) 2010, 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>BoardOpenOnly</code> class represents an implementation of the
 * {@link Board} class that is optimized for fast execution speed when only the
 * Open rule is present in a game of Triple Triad.
 */
public class BoardOpenOnly extends Board
{
    /**
     * The maximum number of cards that can be flipped in a single move when
     * only the Open rule is present.
     */
    private static final int MAX_FLIP_PER_MOVE = 4;

    /**
     * Creates a new <code>BoardOpenOnly</code> in which the two players play
     * with the given hands.
     *
     * @param handOne the starting hand for player one.
     * @param handTwo the starting hand for player two.
     */
    public BoardOpenOnly(Hand handOne, Hand handTwo)
    {
	super(handOne, handTwo, BoardOpenOnly.MAX_FLIP_PER_MOVE);
    }

    /**
     * Computes all of the cards flipped to the current player's colour on a
     * given play.
     *
     * @param space the positional index on the board of the card just played
     * by the current player.
     * @param flipList the array to be filled with the list of flipped board
     * indices.
     * @return the number of cards flipped.
     */
    protected int computeFlips(int space, int[] flipList)
    {
	Card card = this.board[space];
	int nFlips = 0;

	/* Attack up */
        if (space >= 3 && this.board[space - 3] != null &&
            this.amPlayerOne != this.ownedByOne[space - 3] &&
            card.getTop() > this.board[space - 3].getBottom())
	    flipList[nFlips++] = space - 3;

	/* Attack down */
        if (space <= 5 && this.board[space + 3] != null &&
            this.amPlayerOne != this.ownedByOne[space + 3] &&
            card.getBottom() > this.board[space + 3].getTop())
	    flipList[nFlips++] = space + 3;

	/* Attack left */
        if (space % 3 != 0 && this.board[space - 1] != null &&
            this.amPlayerOne != this.ownedByOne[space - 1] &&
            card.getLeft() > this.board[space - 1].getRight())
            flipList[nFlips++] = space - 1;

	/* Attack right */
        if (space % 3 != 2 && this.board[space + 1] != null &&
            this.amPlayerOne != this.ownedByOne[space + 1] &&
            card.getRight() > this.board[space + 1].getLeft())
            flipList[nFlips++] = space + 1;

	return nFlips;
    }
}