/*
 * Copyright (c) 2010, 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>Card</code> class represents a single card in a game of Triple
 * Triad.
 */
public class Card implements Comparable<Card>
{
    /**
     * The number of distinct cards in Final Fantasy 8.
     */
    public static final int NUM_CARDS = 110;

    /**
     * The number of cards per level.
     */
    public static final int CARDS_PER_LEVEL = 11;

    /**
     * The minimal level that a card can be.
     */
    public static final int MIN_LEVEL = 1;

    /**
     * The maximal level that a card can be.
     */
    public static final int MAX_LEVEL = 10;

    /**
     * The minimum value that a card edge can take.
     */
    public static final int MIN_VALUE = 1;

    /**
     * The maximum value that a card edge can take.
     */
    public static final int MAX_VALUE = 10;

    /**
     * The set of all legal cards in Final Fantasy 8.
     */
    private static final Card[] ALL_CARDS = Card.initializeAllCards();

    /**
     * The name of the card.
     */
    private String name;

    /**
     * The index in Final Fantasy 8's list of cards (in PSX memory) at which
     * this card is located.
     */
    private int index;

    /**
     * The value on the top edge of the card.
     */
    private int top;

    /**
     * The value on the right edge of the card.
     */
    private int right;

    /**
     * The value on the bottom edge of the card.
     */
    private int bottom;

    /**
     * The value on the left edge of the card.
     */
    private int left;

    /**
     * The sum of the edge values.
     */
    private int sum;

    /**
     * The elemental attribute of the card.
     */
    private Element element;

    /**
     * Creates a new <code>Card</code> with the four given edge values.
     *
     * @param name the name of the card.
     * @param index the index of the card in memory.
     * @param top the value on the top edge of the card.
     * @param right the value on the right edge of the card.
     * @param bottom the value on the bottom edge of the card.
     * @param left the value on the left edge of the card.
     * @param element the elemental attribute of the card.
     * @throws IllegalArgumentException if any value is outside the range of
     * {@link #MIN_VALUE} and {@link #MAX_VALUE}, or if an invalid index is
     * given.
     */
    private Card(String name, int index, int top, int right, int bottom,
		 int left, Element element)
    {
	if (top < Card.MIN_VALUE || top > Card.MAX_VALUE ||
	    right < Card.MIN_VALUE || right > Card.MAX_VALUE ||
	    bottom < Card.MIN_VALUE || bottom > Card.MAX_VALUE ||
	    left < Card.MIN_VALUE || left > Card.MAX_VALUE)
	    throw new IllegalArgumentException("Invalid card edge value");
	if (index < 0 || index >= Card.NUM_CARDS)
	    throw new IllegalArgumentException("Invalid card index");

	this.name = name;
	this.index = index;
	this.top = top;
	this.right = right;
	this.bottom = bottom;
	this.left = left;
	this.sum = top + right + bottom + left;
	this.element = element;
    }

    /**
     * Returns the name of the card.
     *
     * @return the name of the card.
     */
    public String getName()
    {
	return this.name;
    }

    /**
     * Returns the index of the card in Final Fantasy 8's internal table of
     * cards (in PSX memory).
     *
     * @return the card's index.
     */
    public int getIndex()
    {
	return this.index;
    }

    /**
     * Returns the level of the card.
     *
     * @return the level of the card.
     */
    public int getLevel()
    {
	return (this.index / Card.CARDS_PER_LEVEL) + 1;
    }

    /**
     * Returns the value of the top edge of the card.
     *
     * @return the top edge of the card.
     */
    public int getTop()
    {
	return this.top;
    }

    /**
     * Returns the value of the right edge of the card.
     *
     * @return the right edge of the card.
     */
    public int getRight()
    {
	return this.right;
    }

    /**
     * Returns the value of the bottom edge of the card.
     *
     * @return the bottom edge of the card.
     */
    public int getBottom()
    {
	return this.bottom;
    }

    /**
     * Returns the value of the left edge of the card.
     *
     * @return the left edge of the card.
     */
    public int getLeft()
    {
	return this.left;
    }

    /**
     * Returns the sum of the edge values.
     *
     * @return the sum of the edge values.
     */
    public int getEdgeSum()
    {
	return this.sum;
    }

    /**
     * Returns the minimal edge value of this card.
     *
     * @return the smallest value on any edge.
     */
    public int getSmallestEdge()
    {
	int smallest = this.top;
	if (this.right < smallest)
	    smallest = this.right;
	if (this.bottom < smallest)
	    smallest = this.bottom;
	if (this.left < smallest)
	    smallest = this.left;
	return smallest;
    }

    /**
     * Returns the maximal edge value of this card.
     *
     * @return the maximum value on any edge.
     */
    public int getLargestEdge()
    {
	int largest = this.top;
	if (this.right > largest)
	    largest = this.right;
	if (this.bottom > largest)
	    largest = this.bottom;
	if (this.left > largest)
	    largest = this.left;
	return largest;
    }

    /**
     * Returns the elemental attribute of the card.
     *
     * @return the elemental attribute of the card.
     */
    public Element getElement()
    {
	return this.element;
    }

    /**
     * Returns a <code>String</code> representation of the four edges of this
     * card.
     *
     * @return a string representation of this card.
     */
    public String toString()
    {
	return Integer.toString(this.top) + " " + this.right + " " +
	    this.bottom + " " + this.left;
    }

    /**
     * Compares this <code>Card</code> with the specified <code>Card</code> for
     * order, based on their order in Final Fantasy 8's memory.
     *
     * @param other the <code>Card</code> to which to compare this
     * <code>Card</code>.
     * @return a negative integer, zero, or a positive integer as this object
     * is less than, equal to, or greater than the specified object.
     */
    public int compareTo(Card other)
    {
	return this.index - other.index;
    }

    /**
     * Returns a reference to the card with the four given edge values, if it
     * is a legal card in Final Fantasy 8.
     *
     * @param top the value on the top edge of the card.
     * @param right the value on the right edge of the card.
     * @param bottom the value on the bottom edge of the card.
     * @param left the value on the left edge of the card.
     * @throws IllegalArgumentException if the card is not a valid card in
     * Final Fantasy 8.
     */
    public static Card get(int top, int right, int bottom, int left)
    {
	for (int i = 0; i < Card.ALL_CARDS.length; i++) {
	    Card c = Card.ALL_CARDS[i];
	    if (c.top == top &&
		c.right == right &&
		c.bottom == bottom &&
		c.left == left)
		return c;
	}
	throw new IllegalArgumentException("Card does not exist in FF8");
    }

    /**
     * Returns a reference to the card at the given index in the PSX table of
     * Final Fantasy 8 cards.
     *
     * @param index the index of the card to return.
     * @throws IllegalArgumentException if the given index is negative, or
     * greater than or equal to the number of cards.
     */
    public static Card get(int index)
    {
	if (index < 0 || index >= Card.NUM_CARDS)
	    throw new IllegalArgumentException("Invalid card index: " + index);
	return Card.ALL_CARDS[index];
    }

    /**
     * Returns the set of all legal cards in Final Fantasy 8.
     *
     * @return all of the legal cards in the game.
     */
    private static Card[] initializeAllCards()
    {
	Card[] set = new Card[Card.NUM_CARDS];
	set[0] = new Card("Geezard", 0, 1, 4, 1, 5, Element.NONE);
	set[1] = new Card("Fungar", 1, 5, 1, 1, 3, Element.NONE);
	set[2] = new Card("Bite Bug", 2, 1, 3, 3, 5, Element.NONE);
	set[3] = new Card("Red Bat", 3, 6, 1, 1, 2, Element.NONE);
	set[4] = new Card("Blobra", 4, 2, 3, 1, 5, Element.NONE);
	set[5] = new Card("Gayla", 5, 2, 1, 4, 4, Element.THUNDER);
	set[6] = new Card("Gesper", 6, 1, 5, 4, 1, Element.NONE);
	set[7] = new Card("Fastitocalon-F", 7, 3, 5, 2, 1, Element.EARTH);
	set[8] = new Card("Blood Soul", 8, 2, 1, 6, 1, Element.NONE);
	set[9] = new Card("Caterchipillar", 9, 4, 2, 4, 3, Element.NONE);
	set[10] = new Card("Cockatrice", 10, 2, 1, 2, 6, Element.THUNDER);
	set[11] = new Card("Grat", 11, 7, 1, 3, 1, Element.NONE);
	set[12] = new Card("Buel", 12, 6, 2, 2, 3, Element.NONE);
	set[13] = new Card("Mesmerize", 13, 5, 3, 3, 4, Element.NONE);
	set[14] = new Card("Glacial Eye", 14, 6, 1, 4, 3, Element.ICE);
	set[15] = new Card("Belhelmel", 15, 3, 4, 5, 3, Element.NONE);
	set[16] = new Card("Thrustaevis", 16, 5, 3, 2, 5, Element.WIND);
	set[17] = new Card("Anacondaur", 17, 5, 1, 3, 5, Element.POISON);
	set[18] = new Card("Creeps", 18, 5, 2, 5, 2, Element.THUNDER);
	set[19] = new Card("Grendel", 19, 4, 4, 5, 2, Element.THUNDER);
	set[20] = new Card("Jelleye", 20, 3, 2, 1, 7, Element.NONE);
	set[21] = new Card("Grand Mantis", 21, 5, 2, 5, 3, Element.NONE);
	set[22] = new Card("Forbidden", 22, 6, 6, 3, 2, Element.NONE);
	set[23] = new Card("Armadodo", 23, 6, 3, 1, 6, Element.EARTH);
	set[24] = new Card("Tri-Face", 24, 3, 5, 5, 5, Element.POISON);
	set[25] = new Card("Fastitcalon", 25, 7, 5, 1, 3, Element.EARTH);
	set[26] = new Card("Snow Lion", 26, 7, 1, 5, 3, Element.ICE);
	set[27] = new Card("Ochu", 27, 5, 6, 3, 3, Element.NONE);
	set[28] = new Card("SAM08G", 28, 5, 6, 2, 4, Element.FIRE);
	set[29] = new Card("Death Claw", 29, 4, 4, 7, 2, Element.FIRE);
	set[30] = new Card("Cactaur", 30, 6, 2, 6, 3, Element.NONE);
	set[31] = new Card("Tonberry", 31, 3, 6, 4, 4, Element.NONE);
	set[32] = new Card("Abyss Worm", 32, 7, 2, 3, 5, Element.EARTH);
	set[33] = new Card("Turtapod", 33, 2, 3, 6, 7, Element.NONE);
	set[34] = new Card("Vysage", 34, 6, 5, 4, 5, Element.NONE);
	set[35] = new Card("T-Rexaur", 35, 4, 6, 2, 7, Element.NONE);
	set[36] = new Card("Bomb", 36, 2, 7, 6, 3, Element.FIRE);
	set[37] = new Card("Blitz", 37, 1, 6, 4, 7, Element.THUNDER);
	set[38] = new Card("Wendigo", 38, 7, 3, 1, 6, Element.NONE);
	set[39] = new Card("Torama", 39, 7, 4, 4, 4, Element.NONE);
	set[40] = new Card("Imp", 40, 3, 7, 3, 6, Element.NONE);
	set[41] = new Card("Blue Dragon", 41, 6, 2, 7, 3, Element.POISON);
	set[42] = new Card("Adamantiose", 42, 4, 5, 5, 6, Element.EARTH);
	set[43] = new Card("Hexadragon", 43, 7, 5, 4, 3, Element.FIRE);
	set[44] = new Card("Iron Giant", 44, 6, 5, 6, 5, Element.NONE);
	set[45] = new Card("Behemoth", 45, 3, 6, 5, 7, Element.NONE);
	set[46] = new Card("Chimera", 46, 7, 6, 5, 3, Element.WATER);
	set[47] = new Card("PuPu", 47, 3, 10, 2, 1, Element.NONE);
	set[48] = new Card("Elastiod", 48, 6, 2, 6, 7, Element.NONE);
	set[49] = new Card("GIM47N", 49, 5, 5, 7, 4, Element.NONE);
	set[50] = new Card("Malboro", 50, 7, 7, 4, 2, Element.POISON);
	set[51] = new Card("Ruby Dragon", 51, 7, 2, 7, 4, Element.FIRE);
	set[52] = new Card("Elnoyle", 52, 5, 3, 7, 6, Element.NONE);
	set[53] = new Card("Tonberry King", 53, 4, 6, 7, 4, Element.NONE);
	set[54] = new Card("Wedge, Biggs", 54, 6, 6, 2, 7, Element.NONE);
	set[55] = new Card("Fujin, Raijin", 55, 2, 8, 8, 4, Element.NONE);
	set[56] = new Card("Elvoret", 56, 7, 8, 3, 4, Element.WIND);
	set[57] = new Card("X-ATM092", 57, 4, 8, 7, 3, Element.NONE);
	set[58] = new Card("Granaldo", 58, 7, 2, 8, 5, Element.NONE);
	set[59] = new Card("Gerogero", 59, 1, 8, 8, 3, Element.POISON);
	set[60] = new Card("Iguion", 60, 8, 2, 8, 2, Element.NONE);
	set[61] = new Card("Abadon", 61, 6, 8, 4, 5, Element.NONE);
	set[62] = new Card("Trauma", 62, 4, 8, 5, 6, Element.NONE);
	set[63] = new Card("Oilboyle", 63, 1, 8, 4, 8, Element.NONE);
	set[64] = new Card("Shumi Tribe", 64, 6, 5, 8, 4, Element.NONE);
	set[65] = new Card("Krysta", 65, 7, 5, 8, 1, Element.NONE);
	set[66] = new Card("Propagator", 66, 8, 4, 4, 8, Element.NONE);
	set[67] = new Card("Jumbo Cactaur", 67, 8, 8, 4, 4, Element.NONE);
	set[68] = new Card("Tri-Point", 68, 8, 5, 2, 8, Element.THUNDER);
	set[69] = new Card("Gargantua", 69, 5, 6, 6, 8, Element.NONE);
	set[70] = new Card("Mobile Type 8", 70, 8, 6, 7, 3, Element.NONE);
	set[71] = new Card("Sphinxara", 71, 8, 3, 5, 8, Element.NONE);
	set[72] = new Card("Tiamat", 72, 8, 8, 5, 4, Element.NONE);
	set[73] = new Card("BGH251F2", 73, 5, 7, 8, 5, Element.NONE);
	set[74] = new Card("Red Giant", 74, 6, 8, 4, 7, Element.NONE);
	set[75] = new Card("Catoblepas", 75, 1, 8, 7, 7, Element.NONE);
	set[76] = new Card("Ultima Weapon", 76, 7, 7, 2, 8, Element.NONE);
	set[77] = new Card("Chubby Chocobo", 77, 4, 4, 8, 9, Element.NONE);
	set[78] = new Card("Angelo", 78, 9, 6, 7, 3, Element.NONE);
	set[79] = new Card("Gilgamesh", 79, 3, 7, 9, 6, Element.NONE);
	set[80] = new Card("MiniMog", 80, 9, 3, 9, 2, Element.NONE);
	set[81] = new Card("Chicobo", 81, 9, 4, 8, 4, Element.NONE);
	set[82] = new Card("Quezacotl", 82, 2, 9, 9, 4, Element.THUNDER);
	set[83] = new Card("Shiva", 83, 6, 7, 4, 9, Element.ICE);
	set[84] = new Card("Ifrit", 84, 9, 6, 2, 8, Element.FIRE);
	set[85] = new Card("Siren", 85, 8, 9, 6, 2, Element.NONE);
	set[86] = new Card("Sacred", 86, 5, 1, 9, 9, Element.EARTH);
	set[87] = new Card("Minotaur", 87, 9, 5, 2, 9, Element.EARTH);
	set[88] = new Card("Carbuncle", 88, 8, 4, 10, 4, Element.NONE);
	set[89] = new Card("Diablos", 89, 5, 10, 8, 3, Element.NONE);
	set[90] = new Card("Leviathan", 90, 7, 10, 1, 7, Element.WATER);
	set[91] = new Card("Odin", 91, 8, 10, 3, 5, Element.NONE);
	set[92] = new Card("Pandemona", 92, 10, 1, 7, 7, Element.WIND);
	set[93] = new Card("Cerberus", 93, 7, 4, 6, 10, Element.NONE);
	set[94] = new Card("Alexander", 94, 9, 10, 4, 2, Element.HOLY);
	set[95] = new Card("Phoenix", 95, 7, 2, 7, 10, Element.FIRE);
	set[96] = new Card("Bahumut", 96, 10, 8, 2, 6, Element.NONE);
	set[97] = new Card("Doomtrain", 97, 3, 1, 10, 10, Element.POISON);
	set[98] = new Card("Eden", 98, 4, 4, 9, 10, Element.NONE);
	set[99] = new Card("Ward", 99, 10, 7, 2, 8, Element.NONE);
	set[100] = new Card("Kiros", 100, 6, 7, 6, 10, Element.NONE);
	set[101] = new Card("Laguna", 101, 5, 10, 3, 9, Element.NONE);
	set[102] = new Card("Selphie", 102, 10, 8, 6, 4, Element.NONE);
	set[103] = new Card("Quistis", 103, 9, 6, 10, 2, Element.NONE);
	set[104] = new Card("Irvine", 104, 2, 6, 9, 10, Element.NONE);
	set[105] = new Card("Zell", 105, 8, 5, 10, 6, Element.NONE);
	set[106] = new Card("Rinoa", 106, 4, 10, 2, 10, Element.NONE);
	set[107] = new Card("Edea", 107, 10, 10, 3, 3, Element.NONE);
	set[108] = new Card("Seifer", 108, 6, 9, 10, 4, Element.NONE);
	set[109] = new Card("Squall", 109, 10, 4, 6, 9, Element.NONE);
	return set;
    }
}