/*
 * Copyright (c) 2010, 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>Game</code> class represents a sequence of hands of Triple Triad
 * that can be played between two {@link Player}s.
 */
public class Game
{
    /**
     * The first player in the game.
     */
    private Player one;

    /**
     * The second player in the game.
     */
    private Player two;

    /**
     * The method of interacting with the user.
     */
    private IOMethod io;

    /**
     * Creates a new <code>Game</code> between the two given players.
     *
     * @param playerOne the first player in the game.
     * @param playerTwo the second player in the game.
     * @param io the input/output method used to communicate with the user.
     */
    public Game(Player playerOne, Player playerTwo, IOMethod io)
    {
	this.one = playerOne;
	this.two = playerTwo;
	this.io = io;
    }

    /**
     * Executes the game of Triple Triad on the given board.
     *
     * @param board the board on which to play the game.
     * @return <code>true</code> if the first player wins the game, otherwise
     * <code>false</code>.
     */
    public boolean play(Board board)
    {
	/* Determine who is playing first */
        boolean playerOneTurn = io.isPlayerOneFirst(this.one, this.two);
	if (playerOneTurn == false)
	    board.invert();

        /*
	 * Play until all spaces are filled or until someone aborts or
	 * concedes
	 */
        for (int turn = 0; turn < Board.NUM_SQUARES; turn++) {
	    /* Get the move from the player */
	    Player player = (playerOneTurn ? this.one : this.two);
	    io.turnBegins(player);
	    Move move = player.getMove(board);

	    /* If it is an abort or concede move, we're done */
	    if (move.isAbort())
		return false;
	    else if (move.isConcede())
		return (playerOneTurn == false);

	    /*
	     * Otherwise, move the appropriate card into play, and move on to
	     * the next player's turn
	     */
	    board.play(move);
            playerOneTurn = !(playerOneTurn);
            board.invert();
        }

        /* Compute the final score */
        if (playerOneTurn == false)
            board.invert();
        int oneScore = board.score();
        int twoScore = 2 * Hand.SIZE - oneScore;
        io.finalScore(this.one, this.two, oneScore, twoScore);
	return (oneScore > twoScore);
    }
}