/*
 * Copyright (c) 2010, 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>Hand</code> class represents the set of cards that a player will
 * table to play a game of Triple Triad.  {@link Card}s appear in descending
 * natural order in a hand (that is, high-ranked cards appear at low indices in
 * a hand).
 */
public class Hand
{
    /**
     * The number of cards in one hand.
     */
    public static final int SIZE = 5;

    /**
     * The cards in the hand, stored in descending natural order.
     */
    private Card[] cards;

    /**
     * Creates a new <code>Hand</code> containing the given cards.
     *
     * @param cards the cards to place in the new hand.  The contents of this
     * array are copied, so it is safe to use and/or modify the array after the
     * constructor is complete.
     * @throws IllegalArgumentException if the number of cards in the array is
     * not equal to {@link #SIZE}.
     */
    public Hand(Card[] cards)
    {
	if (cards.length != Hand.SIZE)
	    throw new IllegalArgumentException("Invalid number of cards");	
	this.cards = new Card[Hand.SIZE];

	for (int onCard = 0; onCard < Hand.SIZE; onCard++) {
	    int maxIndex = onCard;
	    for (int onGiven = onCard+1; onGiven < Hand.SIZE; onGiven++) {
		if (cards[onGiven].compareTo(cards[maxIndex]) > 0)
		    maxIndex = onGiven;
	    }
	    this.cards[onCard] = cards[maxIndex];
	    cards[maxIndex] = cards[onCard];
	}
    }

    /**
     * Returns the card at the given index.  Cards are returned in descending
     * natural order.
     *
     * @param index the index at which to get the card.
     * @return the card at the given index.
     */
    public Card getCard(int index)
    {
	return this.cards[index];
    }

    /**
     * Returns the total edge strength in this hand of cards.
     *
     * @return the sum edge strength.
     */
    public int edgeSum()
    {
	int sum = 0;
        for (int i = 0; i < Hand.SIZE; i++)
            sum += this.cards[i].getEdgeSum();
	return sum;
    }

    /**
     * Tests this hand for equality with another hand.  Two hands are
     * considered equal if the contain the exact same cards, though order is
     * unimportant.
     *
     * @param o the object to test for equality.
     * @return <code>true</code> if this hand is equal to the given object,
     * otherwise <code>false</code>.
     */
    public boolean equals(Object o)
    {
	if ((o instanceof Hand) == false)
	    return false;
	Hand other = (Hand)o;

	for (int i = 0; i < Hand.SIZE; i++) {
	    if (this.cards[i].equals(other.cards[i]) == false)
		return false;
	}
	return true;
    }

    /**
     * Returns a hash code for this <code>Hand</code>.
     *
     * @return a hash code for this <code>Hand</code>.
     */
    public int hashCode()
    {
	int ret = 0;
	for (int i = 0; i < Hand.SIZE; i++) {
	    ret <<= (32 / Hand.SIZE);
	    ret ^= this.cards[i].hashCode();
	}
	return ret;
    }
}