/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>HandBuilder</code> class is responsible for building {@link Hand}s
 * of {@link Card}s for play in Triple Triad.
 */
public class HandBuilder
{
    /**
     * The cards that have been added to the builder thus far.
     */
    private Card[] cards;

    /**
     * The number of cards that have been added thus far.
     */
    private int numCards;

    /**
     * Creates a new, empty <code>HandBuilder</code>.
     */
    public HandBuilder()
    {
	this.cards = new Card[Hand.SIZE];
	this.numCards = 0;
    }

    /**
     * Creates a new <code>HandBuilder</code> that contains all of the cards in
     * the given <code>Hand</code> in the same order as in the given hand.
     *
     * @param hand the hand to copy into this builder.
     */
    public HandBuilder(Hand hand)
    {
	this.cards = new Card[Hand.SIZE];
	this.numCards = Hand.SIZE;
	for (int i = 0; i < this.numCards; i++)
	    this.cards[i] = hand.getCard(i);
    }

    /**
     * Creates a new <code>Hand</code> as specified by this builder.
     *
     * @return the created hand.
     * @throws IllegalStateException if the required number of cards have not
     * been added to this builder.
     */
    public Hand toHand()
    {
	if (this.numCards != Hand.SIZE)
	    throw new IllegalStateException("Invalid number of cards in hand");
	return new Hand(this.cards);
    }

    /**
     * Returns the number of cards that have been added to this builder.
     *
     * @return the number of cards in the builder.
     */
    public int getNumCards()
    {
	return this.numCards;
    }

    /**
     * Returns the card at the given index in the builder.
     *
     * @param index the index of the card to return.
     * @return the card at the given index.
     * @throws ArrayIndexOutOfBoundsException if the given index is negative,
     * or greater than or equal to the number of cards in the builder.
     */
    public Card getCard(int index)
    {
	if (index < 0 || index >= this.numCards)
	    throw new ArrayIndexOutOfBoundsException("Invalid array index");
	return this.cards[index];
    }

    /**
     * Adds the given card at the next available index.
     *
     * @param card the card to add.
     * @throws ArrayIndexOutOfBoundsException if there are already
     * {@link Hand#SIZE} cards in the builder.
     * @throws IllegalArgumentException if <code>card</code> is
     * <code>null</code>.
     */
    public void add(Card card)
    {
	if (this.numCards == Hand.SIZE)
	    throw new ArrayIndexOutOfBoundsException("Full hand");
	else if (card == null)
	    throw new IllegalArgumentException("Invalid card");
	this.cards[this.numCards++] = card;
    }

    /**
     * Replaces the card at the given index with the given card.
     *
     * @param index the index at which to replace a card.
     * @param card the new card to place at that index.
     * @throws ArrayIndexOutOfBoundsException if the given index is negative,
     * or greater than or equal to the number of cards in the builder.
     * @throws IllegalArgumentException if <code>card</code> is
     * <code>null</code>.
     */
    public void replace(int index, Card card)
    {
	if (index < 0 || index >= this.numCards)
	    throw new ArrayIndexOutOfBoundsException("Invalid card index");
	else if (card == null)
	    throw new IllegalArgumentException("Invalid card");
	this.cards[index] = card;
    }

    /**
     * Adds the card to the hand to build if it improves the contents of the
     * hand.  Specifically, if the builder does not yet contain the full number
     * of cards required for a hand, the given card will be added at the next
     * available index.  Otherwise, if the given card compares as larger than
     * the smallest card currently in the builder, the given card will replace
     * that card.  If the builder contains the full number of cards and the
     * given card compares as smaller to all of them, the given card is not
     * added.
     *
     * @param card the card to potentially add to the builder.
     * @return <code>true</code> if the given card is added, otherwise
     * <code>false</code>.
     * @throws IllegalArgumentException if the given card is <code>null</code>.
     */
    public boolean offer(Card card)
    {
	if (card == null)
	    throw new IllegalArgumentException("Invalid card");

	if (this.numCards < Hand.SIZE) {
	    this.add(card);
	    return true;
	}

	int minIndex = 0;
	for (int i = 1; i < this.numCards; i++) {
	    if (this.cards[i].compareTo(this.cards[minIndex]) < 0)
		minIndex = i;
	}

	if (this.cards[minIndex].compareTo(card) < 0) {
	    this.cards[minIndex] = card;
	    return true;
	}
	else
	    return false;
    }
}