/*
 * Copyright (c) 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>IOMethod</code> interface represents a way by which the Triple
 * Triad helper program can interact with the user.
 */
public interface IOMethod
{
    /**
     * Prompts the user for whether or not they want to start a new game.
     *
     * @param predicted the hand that the Triple Triad Assistant believes that
     * the user will play.
     * @param bFactory the factory containing the current rules that will be
     * used to build a board.
     * @return the hand that the player will play, or <code>null</code> to
     * indicate that they wish to terminate the program.
     */
    public Hand gameStart(Hand predicted, BoardFactory bFactory);

    /**
     * Returns a hand, input by the user, with which the opponent will play.
     *
     * @param opponent the opponent.
     * @return the hand that the opponent will play.
     */
    public Hand getOpponentHand(Player opponent);

    /**
     * Returns which player is going first based on the user's input.
     *
     * @param playerOne player one.
     * @param playerTwo player two.
     * @return <code>true</code> if player one is going first, otherwise
     * <code>false</code>.
     */
    public boolean isPlayerOneFirst(Player playerOne, Player playerTwo);

    /**
     * Notifies the user that the given player's turn is beginning.
     *
     * @param player the player whose turn is beginning.
     */
    public void turnBegins(Player player);

    /**
     * Notifies the user of the final score of the game.
     *
     * @param playerOne player one.
     * @param playerTwo player two.
     * @param scoreOne the score for the first player.
     * @param scoreTwo the score for the second player.
     */
    public void finalScore(Player playerOne, Player playerTwo, int scoreOne,
                           int scoreTwo);

    /**
     * Notifies the user that an AI player is beginning a search to the given
     * depth.
     *
     * @param depth the depth of the search.
     */
    public void aiStartSearch(int depth);

    /**
     * Notifies the user that the AI's search from the previous call to
     * {@link #aiStartSearch} has finished.
     *
     * @param move the move that the AI chose to make.
     * @param aiScore the AI's projected score for itself.
     * @param oppScore the AI's projected score for its opponent.
     * @param evalRes the internal evaluation function result from the search.
     * @param timeMS the time the search took in milliseconds.
     */
    public void aiEndSearch(Move move, int aiScore, int oppScore, int evalRes,
                            long timeMS);

    /**
     * Prompts the user to input the card and position of a play.
     *
     * @param board the board which requires a play for the currently active
     * player.
     * @return the move that the user input, which could also be to concede or
     * abort the game.
     */
    public Move inputManualMove(Board board);

    /**
     * Prompts the user to input the card and position of a play, verifying
     * the given move against a model.
     *
     * @param board the board which requires a play for the currently active
     * player.
     * @param sv the model against which to verify the input move.
     * @return the move that the user input, which could also be to concede or
     * abort the game.
     */
    public Move inputManualMove(Board board, SearchVerifier sv);

    /**
     * Informs the player about the best opponent card to take.
     *
     * @param card the card to take, where <code>null</code> means that it does
     * not matter.
     */
    public void informBest(Card card);
}