/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>Move</code> class represents which card is being played to which
 * position on any given turn, or the action of conceding the match.
 */
public class Move
{
    /**
     * The smallest position possible.
     */
    public static final int MIN_POSITION = 1;

    /**
     * The largest position possible.
     */
    public static final int MAX_POSITION = Board.NUM_SQUARES;

    /**
     * The card being played, or <code>null</code> if the player wishes to
     * abort or concede the match as their move.
     */
    private Card card;

    /**
     * The position the card is being played to, referenced by position
     * number from {@link #MIN_POSITION} to {@link #MAX_POSITION}, or
     * <code>-1</code> if the player wishes to abort or concede the match as
     * their move.
     */
    private int position;

    /**
     * Whether the player wishes to concede the match (as opposed to just abort
     * it).
     */
    private boolean concede;

    /**
     * Creates a new <code>Move</code> representing a move by the given
     * card to the given position.
     *
     * @param card the card being played.
     * @param position the position from {@link #MIN_POSITION} to
     * {@link #MAX_POSITION} into which the card is being played.
     * @throws IllegalArgumentException if the position is outside the legal
     * range or if the given card is <code>null</code>.
     */
    public Move(Card card, int position)
    {
	if (card == null)
	    throw new IllegalArgumentException("Invalid card");
	if (position < Move.MIN_POSITION || position > Move.MAX_POSITION)
	    throw new IllegalArgumentException("Invalid position: " +
					       position);
	this.card = card;
	this.position = position;
    }

    /**
     * Creates a new <code>Move</code> wherein the player aborts or concedes
     * the match as their move.
     *
     * @param concede whether to concede a loss, as opposed to aborting the
     * match.
     */
    public Move(boolean concede)
    {
	this.card = null;
	this.position = -1;
	this.concede = concede;
    }

    /**
     * Returns the card being played in this move, or <code>null</code> if the
     * player wishes to abort or concede the match.
     *
     * @return the card being played.
     */
    public Card getCard()
    {
	return this.card;
    }

    /**
     * Returns the position into which the card is being played, from
     * {@link #MIN_POSITION} to {@link #MAX_POSITION}, or <code>-1</code> if
     * the player wishes to abort or concede the match.
     *
     * @return the position of the play.
     */
    public int getPosition()
    {
	return this.position;
    }

    /**
     * Returns whether or not this is a move to abort the match.
     *
     * @return <code>true</code> if this is a move to abort the match,
     * otherwise <code>false</code>.
     */
    public boolean isAbort()
    {
	return (this.card == null && this.concede == false);
    }

    /**
     * Returns whether or not this is a move to concede the match.
     *
     * @return <code>true</code> if this is a move to concede the match,
     * otherwise <code>false</code>.
     */
    public boolean isConcede()
    {
	return (this.card == null && this.concede == true);
    }
}