/*
 * Copyright (c) 2010, 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;
import java.util.*;

/**
 * The <code>PlayerAI</code> class represents a Triple Triad player that is
 * controlled by an alpha-beta search AI.
 */
public class PlayerAI extends Player
{
    /**
     * A value that is smaller than any possible evaluation function score, and
     * can be sign-flipped without integer overflow.
     */
    private static final int MIN_BOUND = Integer.MIN_VALUE + 10;

    /**
     * A value that is larger than any possible evalation function score.
     */
    private static final int MAX_BOUND = Integer.MAX_VALUE - 10;

    /**
     * The score granted to an AI player for each card owned at the end of a
     * game.  This score must be greater than twice the possible edge sum of
     * any hand.
     */
    private static final int CARD_SCORE = 100;

    /**
     * The depth of search at which to record a best move.
     */
    private int recordDepth;

    /**
     * The best move to play.
     */
    private SearchBranch bestMove;

    /**
     * Our model for what moves we should test making.
     */
    private SearchModel myModel;

    /**
     * Our model of possible moves our opponent might make.
     */
    private SearchModel oppModel;

    /**
     * Creates a new <code>PlayerAI</code> with the given name.
     *
     * @param name the name of the player.
     * @param io the IO method used to communicate with the user.
     */
    public PlayerAI(String name, IOMethod io)
    {
	super(name, io);
	this.myModel = new SearchAll();
	this.oppModel = new SearchAll();
    }

    /**
     * Creates a new <code>PlayerAI</code> with the given name and model of how
     * their opponent plays.
     *
     * @param name the name of the player.
     * @param oppModel the model of the opposing player.
     * @param io the IO method used to communicate with the user.
     */
    public PlayerAI(String name, SearchModel oppModel, IOMethod io)
    {
	super(name, io);
	this.myModel = new SearchAll();
	this.oppModel = oppModel;
    }

    /**
     * Return the move that this player wants to make on the given board.
     *
     * @param board the board on which to make the move.
     * @return the move to make.
     */
    public Move getMove(Board board)
    {
	/* Set up the alpha-beta search */
	long startTime = System.currentTimeMillis();
	this.recordDepth = board.numFreeLocations();
	this.io.aiStartSearch(this.recordDepth);

	/* Perform the search */
	this.bestMove = null;
	int ret = this.alphaBeta(this.recordDepth, PlayerAI.MIN_BOUND,
				 PlayerAI.MAX_BOUND, board, this.myModel,
				 this.oppModel);
	int myScore = (int)Math.round(((double)ret) / PlayerAI.CARD_SCORE);
	if (myScore < 0)
	    myScore = 2 * Hand.SIZE + myScore;
        int otherScore = Hand.SIZE * 2 - myScore;

	/* Print the results and play */
	long totalTime = System.currentTimeMillis() - startTime;
	Card card = board.cardFromSide(this.bestMove.getCard());
	Move move = new Move(card, this.bestMove.getSpace()+1);
	this.io.aiEndSearch(move, myScore, otherScore, ret, totalTime);
	return move;
    }

    /**
     * Performs an alpha-beta search on the given board from its current
     * perspective.
     *
     * @param depth the remaining depth to which to travel.
     * @param alpha the lower bound of the evaluation function.
     * @param beta the upper bound of the evaluation function.
     * @param board the board on which to search.
     * @param playerModel the model of the current player.
     * @param otherModel the model of the other player.
     * @return the evaluation function value at mini-max depth, or an
     * alpha-beta cutoff value.
     */
    private int alphaBeta(int depth, int alpha, int beta, Board board,
			  SearchModel playerModel, SearchModel otherModel)
    {
	/* Evaluate at leaf nodes */
	if (depth == 0)
	    return this.evaluate(board);

	/* Alpha-beta search over all child nodes */
	Iterator<SearchBranch> acts = playerModel.model(board);
	while (acts.hasNext()) {
	    SearchBranch sb = acts.next();
	    board.play(sb.getCard(), sb.getSpace());
	    board.invert();
	    int ret = -this.alphaBeta(depth-1, -beta, -alpha, board,
				      otherModel, playerModel);
	    board.invert();
	    board.undo();
	    if (ret >= beta)
		return ret;
	    if (ret > alpha) {
		alpha = ret;
		if (depth == this.recordDepth)
		    this.bestMove = sb;
	    }
	}

	return alpha;
    }

    /**
     * Returns the evaluation function run on the current board.  The player
     * who currently has the move on the board gets 100 points for each card
     * they control, plus the point value of each edge of a card still in
     * their hand.
     *
     * @param board the board to evaluate.
     * @return the evaluated score of the board.
     */
    private int evaluate(Board board)
    {
	int ret = board.score() * PlayerAI.CARD_SCORE;
	for (int i = 0; i < Board.NUM_SIDE; i++) {
	    Card side = board.cardFromSide(i);
	    if (side != null) {
		ret += side.getEdgeSum();
		break;
	    }
	}
	return ret;
    }
}