/*
 * Copyright (c) 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;
import java.io.*;
import java.util.*;

/**
 * The <code>Preferences</code> class represents the configuration options
 * specified for an execution of the Triple Triad program.
 */
public class Preferences
{
    /**
     * The comment symbol in the preferences file.
     */
    private static final char COMMENT = '#';

    /**
     * The symbol that delimits a setting name from its value.
     */
    private static final char SETTING = '=';

    /**
     * The user's starting hand.
     */
    private Hand startingHand;

    /**
     * Other settings specified in the file.
     */
    private Map<String, String> settings;

    /**
     * The command line flags that the user can specify.
     */
    private List<String> cmdLineFlags;

    /**
     * The setting that is set by each flag.
     */
    private List<String> cmdLineSettings;

    /**
     * The value assigned to each setting when a flag is present.
     */
    private List<Boolean> cmdLineValues;

    /**
     * Creates a new set of <code>Preferences</code> read from the given file.
     *
     * @param filename the name of the preferences file.
     * @throws PreferencesException if there is an error parsing the
     * preferences file.
     */
    public Preferences(String filename) throws PreferencesException
    {
	this.settings = new HashMap<String, String>();
	this.cmdLineFlags = new ArrayList<String>();
	this.cmdLineSettings = new ArrayList<String>();
	this.cmdLineValues = new ArrayList<Boolean>();

	try {
	    this.parseFile(filename);
	}
	catch (IOException ioe) {
	    throw new PreferencesException("Error reading file: " + filename);
	}
    }

    /**
     * Adds a new command line flag, which sets the given setting to
     * <code>true</code> when provided, to the preferences parser.
     *
     * @param flag the flag to add.
     * @param setting the setting that gets set by this flag.
     * @throws IllegalArgumentException if the flag already exists.
     */
    public void addFlag(String flag, String setting)
    {
	this.addFlag(flag, setting, true);
    }

    /**
     * Adds a new command line flag, which sets the given setting to the given
     * value when provided, to the preferences parser.
     *
     * @param flag the flag to add.
     * @param setting the setting that gets set by this flag.
     * @param value the value to which the setting is set by this flag.
     * @throws IllegalArgumentException if the flag already exists.
     */
    public void addFlag(String flag, String setting, boolean value)
    {
	String flagLower = flag.toLowerCase().trim();
	if (this.cmdLineFlags.contains(flagLower))
	    throw new IllegalArgumentException("Flag already added: " + flag);

	this.cmdLineFlags.add(flagLower);
	this.cmdLineSettings.add(setting.toLowerCase().trim());
	this.cmdLineValues.add(value);
    }

    /**
     * Parses the command line arguments, potentially overriding settings
     * specified in the preferences file.
     *
     * @param args the command line arguments passed to the Java VM.
     * @throws PreferencesException if there are unknown arguments on the
     * command line.
     */
    public void parseCommandLine(String[] args) throws PreferencesException
    {
	for (int i = 0; i < args.length; i++) {
	    int index =
		this.cmdLineFlags.indexOf(args[i].toLowerCase().trim());
	    if (index >= 0) {
		boolean value = this.cmdLineValues.get(index);
		String setting = (value ? "1" : "0");
		this.settings.put(this.cmdLineSettings.get(index), setting);
	    }
	    else {
		String usage = "Options:";
		for (String flag : this.cmdLineFlags)
		    usage += "\n    " + flag;
		throw new PreferencesException(usage);
	    }
	}
    }

    /**
     * Returns the starting hand specified for the user.
     *
     * @return the user's starting hand.
     */
    public Hand getStartingHand()
    {
	return this.startingHand;
    }

    /**
     * Returns the string value of a setting, or <code>null</code> if nothing
     * is specified.
     *
     * @param setting the given setting to look up.
     * @return the string value of the setting.
     */
    public String getString(String setting)
    {
	return this.settings.get(setting.toLowerCase().trim());
    }

    /**
     * Returns the boolean value of a setting, or <code>false</code> if nothing
     * is specified.
     *
     * @param setting the given setting to look up.
     * @return the boolean value of the setting.
     */
    public boolean getBoolean(String setting)
    {
	String value = this.settings.get(setting.toLowerCase().trim());
	if (value == null || value.isEmpty())
	    return false;
	int numVal;
	try {
	    numVal = Integer.valueOf(value);
	}
	catch (Exception e) {
	    return true;
	}
	return (numVal != 0);
    }

    /**
     * Returns the integer value of a setting, or <code>-1</code> if nothing is
     * specified.
     *
     * @param setting the given setting to look up.
     * @return the integer value of the setting.
     */
    public int getInteger(String setting)
    {
	String value = this.settings.get(setting.toLowerCase().trim());
	if (value == null || value.isEmpty())
	    return -1;
	try {
	    return Integer.valueOf(value);
	}
	catch (Exception e) {
	    return -1;
	}
    }

    /**
     * Parses the preferences file and loads the settings contained therein.
     *
     * @param filename the filename of the preferences file.
     * @throws IOException if there is an IO error while reading the file.
     * @throws PreferencesException if there is a syntax error in the file.
     */
    private void parseFile(String filename)
	throws IOException, PreferencesException
    {
	HandBuilder hb = new HandBuilder();
	BufferedReader in = new BufferedReader(new FileReader(filename));

	while (true) {
	    String line = in.readLine();
	    if (line == null)
		break;

	    /* Remove comments */
	    int commentIndex = line.indexOf(Preferences.COMMENT);
	    if (commentIndex >= 0)
		line = line.substring(0, commentIndex);
	    line = line.trim();
	    if (line.isEmpty())
		continue;

	    /* Determine if this is a setting or a card */
	    int setIndex = line.indexOf(Preferences.SETTING);
	    if (setIndex < 0)
		this.parseCardLine(line, hb);
	    else {
		String setting =
		    line.substring(0, setIndex).toLowerCase().trim();
		String value = line.substring(setIndex+1).trim();
		this.parseSettingLine(setting, value);
	    }
	}

	if (hb.getNumCards() != Hand.SIZE)
	    throw new PreferencesException("Starting hand not specified in " +
					   "preferences file");
	this.startingHand = hb.toHand();
    }

    /**
     * Parses a line from the preferences file that contains a setting value.
     *
     * @param setting the setting name.
     * @param value the string value of the setting.
     * @throws PreferencesException if the setting is not legal or a duplicate.
     */
    private void parseSettingLine(String setting, String value)
	throws PreferencesException
    {
	if (setting.isEmpty())
	    throw new PreferencesException("Empty setting name in " +
					   "preferences file");
	if (this.settings.containsKey(setting))
	    throw new PreferencesException("Duplicate setting in " +
					   "preferences file: " + setting);
	this.settings.put(setting, value);
    }

    /**
     * Parses a line from the preferences file that contains a card for the
     * starting hand.
     *
     * @param line the line to parse.
     * @param hb the hand builder that creates the starting hand.
     * @throws PreferencesException if the card is not a legal card, or if the
     * hand is full.
     */
    private void parseCardLine(String line, HandBuilder hb)
	throws PreferencesException
    {
	int[] num = new int[4];

	/* Parse the four integers */
	Scanner scanner = new Scanner(line);
	boolean error = false;
	try {
	    for (int i = 0; i < num.length; i++)
		num[i] = scanner.nextInt();
	}
	catch (Exception e) {
	    error = true;
	}
	if (error || scanner.hasNext())
	    throw new PreferencesException("Invalid card line: " + line);

	/* Grab the card */
	Card card;
	try {
	    card = Card.get(num[0], num[1], num[2], num[3]);
	}
	catch (Exception e) {
	    throw new PreferencesException("Card does not exist: " + line);
	}

	/* Add the card to the builder */
	if (hb.getNumCards() == Hand.SIZE)
	    throw new PreferencesException("Too many cards in preferences " +
					   "file");
	hb.add(card);
    }
}