/*
 * Copyright (c) 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;
import java.util.*;

/**
 * The <code>SearchFF8</code> class models potential moves that would be chosen
 * by a fixed-depth alpha-beta search using the game score as the evaluation
 * function with random selection of branches tied by the evaluation function.
 * To the best of our knowledge, that is how the FF8 AI plays.
 */
public class SearchFF8 implements SearchModel
{
    /**
     * A non-inclusive lower bound on the value on the evaluation function.
     */
    private static final int MIN_BOUND = -10;

    /**
     * A non-inclusive upper bound on the value on the evaluation function.
     */
    private static final int MAX_BOUND = 10;

    /**
     * The maximum search depth used by this FF8 character (different card
     * players use different depths).
     */
    private int maxSearchDepth;

    /**
     * All of the moves tied for the best evaluation at the specified depth.
     */
    private List<SearchBranch> bestMoves;

    /**
     * The depth at which to record the potential moves.
     */
    private int recordDepth;

    /**
     * Creates a new <code>SearchFF8</code> model that searches to the given
     * depth.
     *
     * @param maxSearchDepth the maximum search depth for this model.
     * @throws IllegalArgumentException if the given depth is less than or
     * equal to zero.
     */
    public SearchFF8(int maxSearchDepth)
    {
	if (maxSearchDepth <= 0)
	    throw new IllegalArgumentException("Illegal search depth");
	this.maxSearchDepth = maxSearchDepth;
    }

    /**
     * Returns an iteration over all of the possible moves that a given
     * opponent may make in a given board situation.
     *
     * @param board the situation to which to apply the opponent model.
     * @return an iteration of possible moves that opponent might make.
     */
    public Iterator<SearchBranch> model(Board board)
    {
	/* Set up the alpha-beta search */
	this.bestMoves = new ArrayList<SearchBranch>();
	this.recordDepth = board.numFreeLocations();
	if (this.recordDepth > this.maxSearchDepth)
	    this.recordDepth = this.maxSearchDepth;

	/* Perform the search */
	int ret = this.alphaBeta(this.recordDepth, SearchFF8.MIN_BOUND,
				 SearchFF8.MAX_BOUND, board);
	return this.bestMoves.listIterator();
    }

    /**
     * Performs an alpha-beta search on the given board from its current
     * perspective.
     *
     * @param depth the remaining depth to which to travel.
     * @param alpha the lower bound of the evaluation function.
     * @param beta the upper bound of the evaluation function.
     * @param board the board on which to search.
     * @return the evaluation function value at mini-max depth, or an
     * alpha-beta cutoff value.
     */
    private int alphaBeta(int depth, int alpha, int beta, Board board)
    {
	/* Evaluate at leaf nodes */
	if (depth == 0)
	    return board.score();

	/* Alpha-beta search over all child nodes */
	for (int onCard = Board.NUM_SIDE-1; onCard >= 0; onCard--) {
	    if (board.isCardOnSide(onCard) == false)
		continue;
	    for (int onSpace = 0; onSpace < Board.NUM_SQUARES; onSpace++) {
		if (board.isSquareEmpty(onSpace) == false)
		    continue;
		board.play(onCard, onSpace);
		board.invert();
		int ret = -this.alphaBeta(depth-1, -beta, -alpha, board);
		board.invert();
		board.undo();

		/*
		 * Do not cut ties.  We're not sure how, or if, FF8 cuts any
		 * ties.
		 */
		if (ret > beta)
		    return ret;
		if (ret > alpha) {
		    alpha = ret;
		    if (depth == this.recordDepth) {
			this.bestMoves.clear();
			this.bestMoves.add(new SearchBranch(onCard, onSpace));
		    }
		}
		else if (ret == alpha && depth == this.recordDepth)
		    this.bestMoves.add(new SearchBranch(onCard, onSpace));
	    }
	}

	return alpha;
    }
}