/*
 * Copyright (c) 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;
import java.util.*;

/**
 * The <code>SearchVerifier</code> class is responsible for taking a
 * manually-input move and confirming that it conforms to an expected model of
 * how that player should be moving.
 */
public class SearchVerifier
{
    /**
     * The set of expected moves in this situation.
     */
    private List<SearchBranch> moves;

    /**
     * The board on which the verification occurs.
     */
    private Board board;

    /**
     * Creates a new <code>SearchVerifier</code> that will verify that moves
     * conform to the given model playing on the given board.  If the contents
     * of the board change, this <code>SearchVerifier</code> will no longer
     * remain valid.
     *
     * @param model the model for the player's input.
     * @param board the board on which the player is currently choosing a move.
     */
    public SearchVerifier(SearchModel model, Board board)
    {
	this.board = board;
	this.moves = new ArrayList<SearchBranch>();
	Iterator<SearchBranch> modelMoves = model.model(board);
	while (modelMoves.hasNext())
	    this.moves.add(modelMoves.next());
    }

    /**
     * Verifies that the given move is an expected move by this verifier's
     * model and board.
     *
     * @param move the move to verify.
     * @return <code>true</code> if the move was a concession or abort move or
     * was one of the expected moves generated by the model, otherwise
     * <code>false</code>.
     */
    public boolean verify(Move move)
    {
	if (move.isAbort() || move.isConcede())
	    return true;

	/* Convert the move format to the SearchBranch format */
	int spaceIndex = move.getPosition() - 1;
	int cardIndex = -1;
	Card card = move.getCard();
	for (int i = 0; i < Board.NUM_SIDE; i++) {
	    Card side = this.board.cardFromSide(i);
	    if (side != null && side.equals(card)) {
		cardIndex = i;
		break;
	    }
	}

	/* Iterate over all modelled search branches for a match */
	for (SearchBranch sb : this.moves) {
	    if (sb.getCard() == cardIndex && sb.getSpace() == spaceIndex)
		return true;
	}
	return false;
    }
}