/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;
import java.util.*;

/**
 * The <code>SpoilsChooser</code> class is responsible for choosing which cards
 * to take at the successful conclusion of a match.
 */
public abstract class SpoilsChooser
{
    /**
     * The method of interacting with the user.
     */
    private IOMethod io;

    /**
     * Creates a new <code>SpoilsChooser</code>.
     *
     * @param io the method of interacting with the user.
     */
    public SpoilsChooser(IOMethod io)
    {
	this.io = io;
    }

    /**
     * Informs the user of which card to take from the opponent, and returns
     * the updated player hand.
     *
     * @param playerHand the hand the player had prior to the match.
     * @param opponentHand the hand the opponent played.
     * @return the updated player hand.
     */
    public Hand takeCard(Hand playerHand, Hand opponentHand)
    {
	/*
	 * Sort the cards you could take into two sets: those that would change
	 * your current hand (displace), and those that would not change your
	 * current hand (noDisplace), based on their position in memory
	 * (assuming you always take the bottom five cards from your list with
	 * which to play).
	 */
	Card lowIndexCurrent = playerHand.getCard(Hand.SIZE-1);
	List<Card> displace = new ArrayList<Card>();
	List<Card> noDisplace = new ArrayList<Card>();
	for (int i = 0; i < Hand.SIZE; i++) {
	    Card pos = opponentHand.getCard(i);
	    if (pos.compareTo(lowIndexCurrent) <= 0)
		noDisplace.add(pos);
	    else
		displace.add(pos);
	}

	/* If everything is in noDisplace, it doesn't matter what we take */
	if (displace.isEmpty())
	    return this.doTake(playerHand, null);

	/*
	 * Otherwise, we find the "best" card we can take from the set that
	 * will change our hand.  Here, "best" is defined by the abstract
	 * comparator defined below, since having a higher index doesn't
	 * necessarily mean that it's a stronger card (despite the fact that we
	 * always play our top-indexed five cards).
	 */
	Card bestDisplace = null;
	for (Card pos : displace) {
	    if (bestDisplace == null || this.compare(pos, bestDisplace) > 0)
		bestDisplace = pos;
	}

	/*
	 * Now, compare the best card that could displace our current low-index
	 * card to the low-index card itself.  If it is better, take it.
	 */
	if (this.compare(bestDisplace, lowIndexCurrent) > 0)
	    return this.doTake(playerHand, bestDisplace);

	/*
	 * If we reach this point, everything in the displace set is worse than
	 * the lowest-index card you're holding (note that any card equal to
	 * the low-index card would be in noDisplace).  So, try to take
	 * something that doesn't change the hand you're holding.  If you can't
	 * do that, you just have to take bestDisplace.
	 */
	if (noDisplace.isEmpty())
	    return this.doTake(playerHand, bestDisplace);
	else
	    return this.doTake(playerHand, noDisplace.get(0));
    }

    /**
     * Takes the given card from your opponent's deck and adds it to your own.
     *
     * @param hand the original player hand.
     * @param card the card to take, or <code>null</code> if it does not
     * matter.
     * @return the new player hand.
     */
    private Hand doTake(Hand hand, Card card)
    {
	if (card != null) {
	    HandBuilder hb = new HandBuilder(hand);
	    hb.offer(card);
	    hand = hb.toHand();
	}
	this.io.informBest(card);
	return hand;
    }

    /**
     * Compares two cards for order, ranking higher those cards that you would
     * rather have in your hand.
     *
     * @param o1 the first card.
     * @param o2 the second card.
     * @return a negative integer, zero, or a positive integer as the first
     * card is less than, equal to, or greater than the second.
     */
    protected abstract int compare(Card o1, Card o2);
}