/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>SpoilsMaxEdges</code> class represents a method of choosing cards
 * to take in which cards with higher edge sums are considered better, with
 * ties broken first by a higher maximal edge being larger and second by the
 * card's index being higher.
 */
public class SpoilsMaxEdges extends SpoilsChooser
{
    /**
     * Creates a new <code>SpoilsMaxEdges</code>.
     *
     * @param io the method of interacting with the user.
     */
    public SpoilsMaxEdges(IOMethod io)
    {
	super(io);
    }
    
    /**
     * Compares two cards for order, ranking higher those cards that have a
     * higher edge sum and subsequently those with a higher maximal edge, then
     * a higher index.
     *
     * @param o1 the first card.
     * @param o2 the second card.
     * @return a negative integer, zero, or a positive integer as the first
     * card is less than, equal to, or greater than the second.
     */
    protected int compare(Card o1, Card o2)
    {
	int v1 = o1.getEdgeSum();
	int v2 = o2.getEdgeSum();
	if (v1 != v2)
	    return v1 - v2;
	v1 = o1.getLargestEdge();
	v2 = o2.getLargestEdge();
	if (v1 != v2)
	    return v1 - v2;
	return o1.getIndex() - o2.getIndex();
    }
}