/*
 * Copyright (c) 2010, 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.backend;

/**
 * The <code>TriadHelper</code> class is designed to help you beat the
 * Playstation-controlled AI in Final Fantasy VIII at the game of Triple Triad.
 * It creates one player in a Triple Triad game that is AI-controlled, and one
 * player that is controlled by manual input.  By giving the Playstation's
 * movements as the manually controlled player, you will receive move
 * suggestions from the AI.
 */
public class TriadHelper
{
    /**
     * The preferences file for the application.
     */
    private static final String PREFERENCES = "triadHelper.conf";

    /**
     * The search model preference name.
     */
    private static final String MODEL_SEARCH = "model.search";

    /**
     * The model verification preference name.
     */
    private static final String MODEL_VERIFY = "model.verify";

    /**
     * The FF8 model depth preference name.
     */
    private static final String MODEL_FF8_DEPTH = "model.ff8.depth";

    /**
     * The setting value for the default (full search) options.
     */
    private static final String VALUE_FULL = "Full";

    /**
     * The setting value for the specialized FF8 opponent model.
     */
    private static final String VALUE_FF8 = "FF8";

    /**
     * The parsed user preferences.
     */
    private Preferences preferences;

    /**
     * Which search model to use.
     */
    private SearchModel model;

    /**
     * Whether to verify the search model.
     */
    private boolean verify;

    /**
     * Creates a new <code>TriadHelper</code>.
     */
    public TriadHelper()
    {
	try {
	    this.preferences = new Preferences(TriadHelper.PREFERENCES);
	}
	catch (PreferencesException pe) {
	    System.err.println(pe);
	    System.exit(-1);
	}

	/* Default model if none is specified */
	String prefModelName =
	    this.preferences.getString(TriadHelper.MODEL_SEARCH);
	if (prefModelName == null) {
	    System.out.println("No opponent model specified, using default");
	    prefModelName = TriadHelper.VALUE_FULL;
	}

	/* Determine which model is wanted */
	if (prefModelName.equalsIgnoreCase(TriadHelper.VALUE_FF8)) {
	    int prefDepth =
		this.preferences.getInteger(TriadHelper.MODEL_FF8_DEPTH);
	    System.out.println("Opponent model: FF8's alpha-beta, depth " +
			       prefDepth);
	    this.model = new SearchFF8(prefDepth);
	}
	else if (!prefModelName.equalsIgnoreCase(TriadHelper.VALUE_FULL)) {
	    System.out.println("Unrecognized opponent model, using default");
	    prefModelName = TriadHelper.VALUE_FULL;
	}

	/* Fallback to default */
	if (prefModelName.equalsIgnoreCase(TriadHelper.VALUE_FULL)) {
	    System.out.println("Opponent model: full search");
	    this.model = new SearchAll();
	}

	/* Inform whether model-verification is turned on */
	this.verify = this.preferences.getBoolean(TriadHelper.MODEL_VERIFY);
	System.out.println("Model verification: " +
			   (this.verify ? "on" : "off") + "\n");
    }

    /**
     * Returns the set of preferences for the Triad Helper.
     *
     * @return the preferences for the application.
     */
    public Preferences getPreferences()
    {
	return this.preferences;
    }

    /**
     * Executes the Triple Triad assistant.
     *
     * @param io the method of interacting with the user.
     */
    public void run(IOMethod io)
    {
	PlayerAI one = new PlayerAI("Human Player", this.model, io);
	PlayerManual two = new PlayerManual("Final Fantasy", io);
	if (this.verify)
	    two.setVerification(this.model);

	Game game = new Game(one, two, io);
	BoardFactory bFactory = new BoardFactory();
	Hand playerHand = this.preferences.getStartingHand();
	SpoilsChooser sc = new SpoilsMaxEdges(io);

	while (true) {
	    playerHand = io.gameStart(playerHand, bFactory);
	    if (playerHand == null)
		break;
	    Hand ff8Hand = io.getOpponentHand(two);
	    Board board = bFactory.createBoard(playerHand, ff8Hand);
	    if (game.play(board))
		playerHand = sc.takeCard(playerHand, ff8Hand);
	}
    }
}