/*
 * Copyright (c) 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.console;
import java.util.*;

/**
 * The <code>ColumnDisplay</code> class represents a way of nicely formatting
 * columns of different lengths and widths onto the console.
 */
public class ColumnDisplay
{
    /**
     * The default amount of space between columns is none is specified.
     */
    public static final int DEFAULT_SPACE = 8;

    /**
     * The contents of the columns.
     */
    private List<List<String>> columns;

    /**
     * The number of spaces between the columns.
     */
    private int spaces;

    /**
     * Creates a new <code>ColumnDisplay</code> with the given number of
     * columns.
     *
     * @param cols the number of columns.
     * @throws IllegalArgumentException if the number of columns is not
     * positive.
     */
    public ColumnDisplay(int cols)
    {
	this(cols, ColumnDisplay.DEFAULT_SPACE);
    }

    /**
     * Creates a new <code>ColumnDisplay</code> with the given number of
     * columns and the given number of spaces between the columns.
     *
     * @param cols the number of columns.
     * @param spaces the number of spaces between each columns.
     * @throws IllegalArgumentException if the number of columns is not
     * positive, or if the number of spaces is negative.
     */
    public ColumnDisplay(int cols, int spaces)
    {
	if (cols <= 0)
	    throw new IllegalArgumentException("Invalid number of columns");
	if (spaces < 0)
	    throw new IllegalArgumentException("Invalid number of spaces");
	this.columns = new ArrayList<List<String>>();
	for (int i = 0; i < cols; i++)
	    this.columns.add(new ArrayList<String>());
	this.spaces = spaces;
    }

    /**
     * Adds the given output to the given columns.
     *
     * @param column the column, indexed from zero.
     * @param output the output to add to the column.
     * @throws IllegalArgumentException if the output is <code>null</code>.
     * @throws ArrayIndexOutOfBoundsException if the column index is negative,
     * or greater than or equal to the number of columns.
     */
    public void add(int column, String output)
    {
	if (output == null)
	    throw new IllegalArgumentException("Invalid output string");
	this.columns.get(column).add(output);
    }

    /**
     * Returns whether the given column is empty.
     *
     * @param column the column, indexed from zero.
     * @throws ArrayIndexOutOfBoundsException if the column index is negative,
     * or greater than or equal to the number of columns.
     */
    public boolean isColumnEmpty(int column)
    {
	return this.columns.get(column).isEmpty();
    }

    /**
     * Outputs the columns to the console.
     */
    public void display()
    {
	/* Determine the number of rows and columns */
	int cols = this.columns.size();
	int rows = 0;
	for (List<String> column : this.columns) {
	    if (column.size() > rows)
		rows = column.size();
	}

	/* Determine the first and last row on which each column will appear */
	int[] firstRow = new int[cols];
	for (int i = 0; i < cols; i++)
	    firstRow[i] = (rows - this.columns.get(i).size())/2;

	/* Determine the width of each columns */
	int[] width = new int[cols];
	for (int i = 0; i < cols; i++) {
	    width[i] = 0;
	    List<String> contents = this.columns.get(i);
	    for (String str : contents) {
		int len = str.length();
		if (len > width[i])
		    width[i] = len;
	    }
	}

	/* Output the rows */
	for (int onRow = 0; onRow < rows; onRow++) {
	    for (int onCol = 0; onCol < cols; onCol++) {
		List<String> col = this.columns.get(onCol);
		String output = "";
		if (onRow >= firstRow[onCol] &&
		    onRow < firstRow[onCol] + col.size())
		    output = col.get(onRow - firstRow[onCol]);
		StringBuilder sb =
		    new StringBuilder(width[onCol] + this.spaces);
		sb.append(output);
		int toAdd = 0;
		if (onCol != cols - 1)
		    toAdd = width[onCol] - output.length() + this.spaces;
		for (int i = 0; i < toAdd; i++)
		    sb.append(' ');
		System.out.print(sb.toString());
	    }
	    System.out.println();
	}
    }
}