/*
 * Copyright (c) 2010, 2012, 2013 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.console;
import java.io.*;
import java.util.*;
import triad.backend.*;

/**
 * The <code>IOConsole</code> class represents a method of interacting with the
 * Triple Triad Helper by using a keyboard on the console.
 */
public class IOConsole implements IOMethod
{
    /**
     * The buffered reader of standard input, that is, the keyboard.
     */
    private BufferedReader in;

    /**
     * Whether to output extra information, to make this program function like
     * a fully-fledged console game.  If verbose mode is off, output will be
     * streamlined for speedrunning.
     */
    private boolean verbose;

    /**
     * Whether or not the user input a concede command.
     */
    private boolean concedeInput;

    /**
     * Creates a new <code>IOConsole</code> method for interacting with the
     * Triple Triad program and initializes the keyboard buffer.
     *
     * @param verbose whether or not to output extra information.  If verbose
     * mode is on, the program will function like a fully-fledged console game.
     * Otherwise, output will be streamlined for speedrunning.
     */
    public IOConsole(boolean verbose)
    {
        this.in = new BufferedReader(new InputStreamReader(System.in));
        this.verbose = verbose;
    }

    /**
     * Prompts the user for whether or not they want to start a new game.
     *
     * @param predicted the hand that the Triple Triad Assistant believes that
     * the user will play.
     * @param bFactory the factory containing the current rules that will be
     * used to build a board.
     * @return the hand that the player will play, or <code>null</code> to
     * indicate that they wish to terminate the program.
     */
    public Hand gameStart(Hand predicted, BoardFactory bFactory)
    {
	/* Output the predicted deck */
	HandBuilder hb = new HandBuilder(predicted);
	this.printHandRules(hb, bFactory);

	/* Get the user's choice */
	while (true) {
	    System.out.print("(N)ew game | (C)orrect # | (Q)uit: ");
	    System.out.flush();
	    StringTokenizer st = new StringTokenizer(this.readLine());
	    if (st.hasMoreTokens() == false)
		continue;
	    char command = st.nextToken().toLowerCase().charAt(0);
	    if (command == 'n') {
		System.out.println();
		return hb.toHand();
	    }
	    else if (command == 'q')
		return null;
	    else if (command != 'c' || st.hasMoreTokens() == false)
		continue;
	    int cardToChange = -1;
	    try {
		cardToChange = Integer.parseInt(st.nextToken());
	    }
	    catch (Exception ex) {
		continue;
	    }
	    if (cardToChange <= 0 || cardToChange > Hand.SIZE)
		continue;

	    hb.replace(cardToChange-1, this.readCard(false));
	    System.out.println();
	    this.printHandRules(hb, bFactory);
	}
    }

    /**
     * Returns a hand, input by the user, with which the opponent will play.
     *
     * @param opponent the opponent.
     * @return the complete hand for that opponent.
     */
    public Hand getOpponentHand(Player opponent)
    {
        System.out.println("Input a hand for " + opponent);
	HandBuilder hb = new HandBuilder();
        for (int i = 0; i < Hand.SIZE; i++)
            hb.add(this.readCard(false));
        System.out.println();
        return hb.toHand();
    }

    /**
     * Returns which player is going first based on the user's input.
     *
     * @param playerOne player one.
     * @param playerTwo player two.
     * @return <code>true</code> if player one is going first, otherwise
     * <code>false</code>.
     */
    public boolean isPlayerOneFirst(Player playerOne, Player playerTwo)
    {
        boolean f = this.readYesNo("Is " + playerOne + " going first? ");
	System.out.println();
	return f;
    }

    /**
     * Notifies the user that the given player's turn is beginning.
     *
     * @param player the player whose turn is beginning.
     */
    public void turnBegins(Player player)
    {
        System.out.println("Time for " + player + " to move");
    }

    /**
     * Notifies the user of the final score of the game.
     *
     * @param playerOne player one.
     * @param playerTwo player two.
     * @param scoreOne the score for the first player.
     * @param scoreTwo the score for the second player.
     */
    public void finalScore(Player playerOne, Player playerTwo, int scoreOne,
                           int scoreTwo)
    {
        System.out.println("Final score:  " + scoreOne + " - " + scoreTwo);
	if (scoreOne <= scoreTwo)
	    System.out.println();
    }

    /**
     * Notifies the user that an AI player is beginning a search to the given
     * depth.
     *
     * @param depth the depth of the search.
     */
    public void aiStartSearch(int depth)
    {
        System.out.print("Doing depth " + depth + " search...");
        System.out.flush();
    }

    /**
     * Notifies the user that the AI's search from the previous call to
     * {@link #aiStartSearch} has finished.
     *
     * @param move the move that the AI chose to make.
     * @param aiScore the AI's projected score for itself.
     * @param oppScore the AI's projected score for its opponent.
     * @param evalRes the internal evaluation function result from the search.
     * @param timeMS the time the search took in milliseconds.
     */
    public void aiEndSearch(Move move, int aiScore, int oppScore, int evalRes,
                            long timeMS)
    {
        Card card = move.getCard();
        int spot = move.getPosition();

        if (this.verbose) {
            System.out.println(Long.toString(timeMS) + " ms");
            System.out.println("Evaluation = " + evalRes + " (" + aiScore +
                               " - " + oppScore + ")");
            System.out.println("Play card " + card +  " to spot " + spot);
	    System.out.println();
        }
        else
            System.out.println("\nPlay " + card +  " to spot " + spot +
                               " (score " + aiScore + " - " + oppScore +
			       ")\n");
    }

    /**
     * Prompts the user to input the card and position of a play.
     *
     * @param board the board which requires a play for the currently active
     * player.
     * @return the move that the user input, which could also be to concede or
     * abort the game.
     */
    public Move inputManualMove(Board board)
    {
        /* Print available cards */
        if (this.verbose) {
            System.out.println("Available cards:");
	    for (int i = 0; i < Hand.SIZE; i++) {
		Card card = board.cardFromSide(i);
		if (card != null)
		    System.out.println("- " + card);
	    }
        }

        /* Pick a card to play */
	Card card;
        while (true) {
            card = this.readCard(true);
	    if (card == null) {
		System.out.println();
		return new Move(this.concedeInput);
	    }
	    boolean found = false;
            for (int i = 0; i < Hand.SIZE; i++) {
		Card pos = board.cardFromSide(i);
		if (pos != null && pos.equals(card)) {
		    found = true;
		    break;
		}
            }
	    if (found)
		break;
            System.out.println("Card does not match any card in hand");
        }

	Move m = new Move(card, this.inputManualMovePosition(board));
	System.out.println();
	return m;
    }

    /**
     * Prompts the user to input the card and position of a play, verifying
     * the given move against a model.
     *
     * @param board the board which requires a play for the currently active
     * player.
     * @param sv the model against which to verify the input move.
     * @return the move that the user input, which could also be to concede or
     * abort the game.
     */
    public Move inputManualMove(Board board, SearchVerifier sv)
    {
	Move move;
	while (true) {
	    move = this.inputManualMove(board);
	    if (sv.verify(move))
		return move;
	    boolean proceed = this.readYesNo("WARNING: Move does not match " +
					     "model. Proceed? ");
	    System.out.println();
	    if (proceed)
		return move;
	}
    }

    /**
     * Informs the player about the best opponent card to take.
     *
     * @param card the card to take, or <code>null</code> if it does not matter
     * what the player takes.
     */
    public void informBest(Card card)
    {
	if (card != null)
	    System.out.println("Card to take: " + card + "\n");
	else
	    System.out.println("Card to take: anything\n");
    }

    /**
     * Outputs the current set of cards in the hand builder and the current set
     * of rules in the board factory.
     *
     * @param hb the hand builder to output.
     * @param bf the board factory to output.
     */
    private void printHandRules(HandBuilder hb, BoardFactory bf)
    {
	ColumnDisplay cd = new ColumnDisplay(2);
	for (int i = 0; i < Hand.SIZE; i++)
	    cd.add(0, hb.getCard(i).toString());

	List<String> rules = bf.getRules();
	for (String rule : rules)
	    cd.add(1, rule);
	
	cd.display();
    }

    /**
     * Prompts the user to input the position to which they want to play their
     * previously selected card.
     *
     * @param board the board on which the user is currently the active (as
     * opposed to inverted) player.
     * @return a position in the range of 1 to 9.
     */
    private int inputManualMovePosition(Board board)
    {
        /* Display available positions */
        if (this.verbose) {
            System.out.println("Available locations:");
	    String display = "";
	    for (int i = 0; i < Board.NUM_SQUARES; i++) {
		if (i % 3 != 0)
		    display += " ";
		if (board.isSquareEmpty(i))
		    display += (i + 1);
		else
		    display += "-";
		if (i % 3 == 2 && i != Board.NUM_SQUARES - 1)
		    display += "\n";
	    }
            System.out.println(display);
        }
	
        /* Get a position to play */
        int posToPlay = -1;
        while (posToPlay == -1) {
            posToPlay = this.readInteger("Choose a location: ");
            if (posToPlay <= 0 || posToPlay > Board.NUM_SQUARES) {
                System.out.println("Invalid location");
                posToPlay = -1;
	    }
	    else if (board.isSquareEmpty(posToPlay-1) == false) {
		System.out.println("Position already occupied");
		posToPlay = -1;
            }
        }
        return posToPlay;
    }

    /**
     * Prompts the user to input a valid card from Final Fantasy 8.
     *
     * @param allowAbortConcede whether or not to allow an abort or concede
     * command to be input.
     * @return the card input, or <code>null</code> if an abort or concede is
     * input.
     */
    private Card readCard(boolean allowAbortConcede)
    {
	int[] num = new int[4];
	String prompt;
	if (allowAbortConcede == false)
	    prompt = "Enter card: ";
	else
	    prompt = "(A)bort | (C)oncede | Enter card: ";

	while (true) {
	    /* Output the prompt and check for abort/concede input */
	    System.out.print(prompt);
	    System.out.flush();
	    String line = this.readLine().trim();
	    if (line.isEmpty())
		continue;
	    char command = line.toLowerCase().charAt(0);
	    if (allowAbortConcede && command == 'a') {
		this.concedeInput = false;
		return null;
	    }
	    else if (allowAbortConcede && command == 'c') {
		this.concedeInput = true;
		return null;
	    }

	    /* Integer input: parse the input and return the card */
	    Scanner scanner = new Scanner(line);
	    try {
		for (int i = 0; i < num.length; i++)
		    num[i] = scanner.nextInt();
	    }
	    catch (Exception e) {
                System.out.println("Invalid integers");
		continue;
            }
	    if (scanner.hasNext()) {
		System.out.println("Too many numbers input");
		continue;
	    }
	    try {
                return Card.get(num[0], num[1], num[2], num[3]);
            }
            catch (Exception e) {
                System.out.println("Card does not exist in Final Fantasy 8");
            }
	}
    }

    /**
     * Reads a single integer from a single line.
     *
     * @param prompt the prompt to display before the user types.
     * @return the parsed integer.
     */
    private int readInteger(String prompt)
    {
        int ret = 0;
        while (true) {
            System.out.print(prompt);
            System.out.flush();
            try {
                String line = this.readLine();
                line = line.trim();
                ret = Integer.parseInt(line);
                break;
            }
            catch (Exception e) {
                System.out.println("Invalid integer");
            }
        }
        return ret;
    }

    /**
     * Reads whether a line starts with a <code>y</code> or </code>Y</code>,
     * or a <code>n</code> or </code>N</code>.
     *
     * @param prompt the prompt to display before the user types.
     * @return the parsed boolean value.
     */
    private boolean readYesNo(String prompt)
    {
        boolean ret = false;
        while (true) {
            System.out.print(prompt);
            System.out.flush();
            try {
                String line = this.readLine();
                line = line.trim();
                char c = line.charAt(0);
                if (c == 'Y' || c == 'y') {
                    ret = true;
                    break;
                }
		else if (c == 'N' || c == 'n') {
                    ret = false;
                    break;
                }
            }
            catch (Exception e) {}
        }
        return ret;
    }

    /**
     * Reads a line of input from the console and returns it.
     *
     * @return the line of input read.
     */
    private String readLine()
    {
	String line = null;
	while (line == null) {
	    try {
		line = this.in.readLine();
	    }
	    catch (IOException ioe) {
		System.out.println("Error reading line from standard input");
	    }
	}
	return line;
    }
}