/*
 * Copyright (c) 2011, 2012 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package triad.console;
import java.io.*;
import java.util.*;
import triad.backend.*;

/**
 * The <code>ProcessDump</code> class is responsible for reading in a recording
 * of the cards played by a given NPC in Final Fantasy 8, over a large number
 * of matches, and outputting statistic information about those matches.
 */
public class ProcessDump
{
    /**
     * Runs the <code>ProcessDump</code> command by reading in card dumps off
     * of standard input.
     *
     * @param args the arguments to the Java VM, which are ignored.
     * @throws IOException if there is an error reading standard input.
     */
    public static void main(String[] args) throws IOException
    {
	int totalHands = 0;
	int handsWithTop = 0;
	double averageStr = 0.0;
	Set<Hand> uniqueHands = new HashSet<Hand>();
	Set<Card> uniqueCards = new HashSet<Card>();
	boolean[] levelPlayed = new boolean[Card.MAX_LEVEL];
	for (int i = 0; i < levelPlayed.length; i++)
	    levelPlayed[i] = false;

	/* Read in all of the hands played, one per line */
	BufferedReader in =
	    new BufferedReader(new InputStreamReader(System.in));
	while (true) {
	    Hand hand = ProcessDump.readHand(in);
	    if (hand == null)
		break;

	    /* Card statistics */
	    boolean hasTop = false;
	    for (int i = 0; i < Hand.SIZE; i++) {
		Card card = hand.getCard(i);
		int level = card.getLevel();
		uniqueCards.add(card);
		levelPlayed[level-1] = true;
		if (level == Card.MAX_LEVEL)
		    hasTop = true;
	    }

	    /* Hand statistics */
	    totalHands++;
	    uniqueHands.add(hand);
	    averageStr += hand.edgeSum();
	    if (hasTop)
		handsWithTop++;
	}
	averageStr /= totalHands * Hand.SIZE * 4;

	/* Output the statistics about all of the hands played */
	System.out.println("Total hands processed:         " + totalHands);
	System.out.println("Unique unordered hands played: " +
			   uniqueHands.size());
	System.out.println("Unique cards played:           " +
			   uniqueCards.size());
	System.out.format("Hands with top-level card:     " + handsWithTop +
			  " (%.2f%%)%n",
			  ((double)handsWithTop) / totalHands * 100);
	System.out.format("Average card strength:         %.2f%n", averageStr);
	System.out.println("Levels of cards played:        " +
			   ProcessDump.levelsPlayed(levelPlayed));
    }

    /**
     * Reads the next hand from the buffered reader.
     *
     * @param in the buffered reader to read.
     * @return the next hand, or <code>null</code> if the end of the input has
     * been reached.
     * @throws IOException if there is an error reading from the buffered
     * reader.
     */
    private static Hand readHand(BufferedReader in) throws IOException
    {
	/* Read the next input line */
	String line;
	while (true) {
	    line = in.readLine();
	    if (line == null)
		return null;
	    line = line.trim();
	    if (line.isEmpty() == false)
		break;
	}

	/* Parse the cards from the line */
	Scanner scanner = new Scanner(line);
	HandBuilder hb = new HandBuilder();
	for (int i = 0; i < Hand.SIZE; i++)
	    hb.add(Card.get(scanner.nextInt()));
	if (scanner.hasNext())
	    throw new IllegalArgumentException("Invalid hand line: " + line);
	return hb.toHand();
    }

    /**
     * Converts the boolean list of levels played to a string representation.
     *
     * @param played the boolean flags for which levels were played.
     * @return a string representation of the levels played.
     */
    private static String levelsPlayed(boolean[] played)
    {
	String ret = "";
	boolean needComma = false;
	for (int i = 0; i < played.length; i++) {
	    if (played[i]) {
		if (needComma)
		    ret += ", " + (i+1);
		else {
		    ret += (i+1);
		    needComma = true;
		}
	    }
	}
	return ret;
    }
}